/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import java.util.Map;
import java.util.Set;

/**
 * Base class for all aspect dependent messages. Subclasses must be immutable so that the same class
 * instance can be safely passed to all listeners.
 *
 * @author Sergey Groshev
 */
public abstract class AspectMessage {

  private final String aspect;
  private final AspectAdapter adapter;
  private final String channel;

  protected AspectMessage(AspectAdapter adapter) {
    this(adapter.getAspect(), adapter.getCurrentChannel(), adapter);
  }

  protected AspectMessage(String aspect, AspectAdapter adapter) {
    this(aspect, null, adapter);
  }

  protected AspectMessage(String aspect, String channel, AspectAdapter adapter) {
    this.aspect = aspect;
    this.channel = channel;
    this.adapter = adapter;
  }

  public final String getAspect() {
    return aspect;
  }

  public final String getChannel() {
    return channel;
  }

  public final AspectAdapter getAdapter() {
    return adapter;
  }

  /**
   * Gets active tags from provider.
   * == {@link TraceProvider#getAllTags()}
   *
   * @return Set of tags copy. Changes have no effects on trace handling.
   */
  public final Map<String, Set<String>> getAllTags() {
    if (adapter == null) {
      return null;
    }
    return adapter.getAllTags();
  }

  /**
   * Gets tags for the current message from provider.
   * == {@link TraceProvider#getTags}
   *
   * @return Set of tags which are part of provider state.
   */
  protected final Set<String> getMyTags() {
    if (adapter == null) {
      return null;
    }
    return adapter.getMyTags();
  }
}
