/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import com.unitesk.aspectrace.provider.StreamTraceProvider;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * A source of stream data represented by {@link AbstractTraceMessage raw messages}. Depending on
 * realization may or may not have the beginning and end, work either in model PUSH or PULL.
 * Messages that appear in message stream are being send to all connected {@link
 * MessageStreamListener listeners}.
 *
 * <p>May be used as common data bus of application or application group. Can be converted to {@link
 * TraceProvider aspect message trace} via {@link StreamTraceProvider}.
 *
 * <p>Note: Realizations of methods {@link #fireStreamStart()}, {@link
 * #fireStreamMessage(AbstractTraceMessage)} and {@link #fireStreamEnd(boolean)} are made {@code
 * synchronized}, but not {@code final}.
 *
 * <p>@author Sergey Groshev
 */
public abstract class MessageStream {

  // TODO: filtration via MessageStreamFilter.
  // To insert directly in the stream or make MessageStream cascade.

  protected List<MessageStreamListener> listeners = new ArrayList<MessageStreamListener>();

  protected MessageStream() {}

  /** Adds listener to the list. */
  public void addListener(MessageStreamListener msl) {
    if (msl == null) {
      throw new NullPointerException();
    }
    listeners.add(msl);
  }

  /** Adds listener to the list. */
  public void removeListener(MessageStreamListener msl) {
    listeners.remove(msl);
  }

  /**
   * Begin message stream processing loop in case of an active stream working in model PULL. In this
   * case the realization must call method {@link #fireStreamStart()} at the beginning of processing
   * then extract messages from it's source and call for each of them method {@link
   * #fireStreamMessage(AbstractTraceMessage)}, and in case of stream end (stream may never end) -
   * method {@link #fireStreamEnd(boolean)} with the flag marking either a normal end or abort.
   * Method returns when all messages are extracted and processed or when the message stream closure
   * is not normal.
   *
   * <p>The method realization is empty by default which corresponds with the model PUSH. For the
   * streams working in this model and processing messages from other active sources the
   * corresponding calls must be done in the same order in the corresponding listeners.
   *
   * @throws IOException Input-Output error.
   * @throws IllegalStateException Processing has already begun.
   * @throws IllegalArgumentException Error in stream.
   */
  public void process() throws IOException, IllegalStateException, IllegalArgumentException {}

  /**
   * Sends stream begin message to all listeners. Exceptions from the listeners are being ignored.
   */
  protected synchronized void fireStreamStart() {
    for (MessageStreamListener msl : listeners) {
      try {
        msl.handleStreamStart();
      } catch (Throwable e) {
        // Silently ignore
      }
    }
  }

  /** Sends message to all listeners. Exceptions from the listeners are being ignored. */
  protected synchronized void fireStreamMessage(AbstractTraceMessage msg) {
    // TODO: filters
    for (MessageStreamListener msl : listeners) {
      try {
        msl.handleStreamMessage(msg);
      } catch (Throwable e) {
        // Silently ignore
      }
    }
  }

  /** Sends stream end message to all listeners. Exceptions from the listeners are being ignored. */
  protected synchronized void fireStreamEnd(boolean normal) {
    for (MessageStreamListener msl : listeners) {
      try {
        msl.handleStreamEnd(normal);
      } catch (Throwable e) {
        // Silently ignore
      }
    }
  }
}
