/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Class representing call parameter of method, oracle, scenario etc..
 *
 * @author Sergey Groshev
 */
public class Parameter {

  protected String type;
  protected String name;
  protected String value;

  /**
   * Constructor.
   *
   * @param type Parameter type.
   * @param name Parameter name.
   * @param value Parameter value.
   */
  public Parameter(String type, String name, Object value) {
    super();
    this.type = type;
    this.name = name;
    this.value = String.valueOf(value);
  }

  public String getType() {
    return type;
  }

  public String getName() {
    return name;
  }

  public String getValue() {
    return value;
  }

  @Override
  public String toString() {
    return "Parameter " + (type == null ? "" : type) + ' ' + name + '=' + value;
  }

  // ------------------------------------------------------
  // Converters to and from TraceMessage

  /** Converts to a node for the more convenient packing into trace messages. */
  public TraceNode toNode() {
    Map<String, String> attrs = new HashMap<String, String>(2, 1.0f);
    if (type != null) {
      attrs.put("type", type);
    }
    if (name != null) {
      attrs.put("name", name);
    }
    TraceNode node = new TraceNode("param", attrs);
    if (value != null) {
      node.addTextChild(value);
    }
    return node;
  }

  /** Restores a parameter from the corresponding trace message part. */
  public static Parameter fromNode(TraceNode tn) throws IllegalArgumentException {
    if (tn.name != "param") {
      throw new IllegalArgumentException("Parameter expected");
    }
    List<TraceNode> children = tn.getChildren();
    String value;
    switch (children.size()) {
      case 1:
        TraceNode child = children.get(0);
        if (child.isText()) {
          value = child.getText();
          break;
        }
        // else fall through
      default:
        throw new IllegalArgumentException("Unexpected child of parameter node");
      case 0:
        value = null;
    }

    return new Parameter(tn.getAttr("type"), tn.getAttr("name"), value);
  }
}
