/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import java.util.List;
import java.util.Map;

/**
 * Trace message. It is passed to a tracer to write in a trace; after the trace has been processed
 * returns exactly as it was before then is converted by a corresponding adapter into the
 * aspect-dependent messages for use by handlers.
 *
 * <p>It is possible for an aspect tracer to create trace messages which are used for a direct
 * passing to the corresponding aspect adapter, without any intermediate forms; this messages can be
 * passed directly to handlers and be converted to a tree of {@link TraceNode} only in case it is
 * needed to write in an external form.
 *
 * <p>If a message is formed to be directly passed to an adapter then the first query to the
 * properties of the node (<code>name, attrs, children</code>) triggers a conversion to a tree of
 * {@link TraceNode}.
 *
 * @author Sergey Groshev
 */
public final class TraceMessage extends AbstractTraceMessage {
  private TraceNode node;
  private LazyNode lazyNode;

  /** Constructor for {@link TraceProvider trace sources}. */
  public TraceMessage(String aspect, String channel, TraceNode node) {
    super(aspect, channel);
    this.node = node;
    this.lazyNode = null;
    node.setReady();
  }

  /** Constructor for tracing. */
  public TraceMessage(String aspect, TraceNode node) {
    this(aspect, null, node);
  }

  /** Constructor for a direct data pass to handlers. */
  public TraceMessage(String aspect, String channel, LazyNode node) {
    super(aspect, channel);
    this.lazyNode = node;
    this.node = null;
  }

  /** Constructor for a direct data pass to handlers. */
  public TraceMessage(String aspect, LazyNode node) {
    this(aspect, null, node);
  }

  @Override
  public Kind getKind() {
    return Kind.TRACE_MESSAGE;
  }

  /** Gets a root node of message. The node is created using a delayed conversion if needed. */
  public TraceNode getNode() {
    if (node == null) {
      node = getLazyNode().toNode();
    }
    return node;
  }

  /**
   * Gets raw data from the aspect library if this data is given. This data maybe converted to a
   * normal tree of {@link TraceNode узлов} if necessary.
   */
  public LazyNode getLazyNode() {
    return lazyNode;
  }

  // -----------------------------------------
  // Shortcuts to root node

  /**
   * Gets message type name. Names and the corresponding message structures are determined by aspect
   * data model.
   *
   * <p>The name is requested from the root node. {@code == getNode().getName()}
   */
  public String getName() {
    return getNode().getName();
  }

  /** Get root node attributes. {@code == getNode().getAttrs()} */
  public Map<String, String> getAttrs() {
    return getNode().getAttrs();
  }

  /** Get root node attribute value. {@code == getNode().getAttr(attrName)} */
  public String getAttr(String attrName) {
    return getNode().getAttr(attrName);
  }

  /** Remove root node attribute. {@code == getNode().removeAttr(attrName)} */
  public void removeAttr(String attrName) {
    getNode().removeAttr(attrName);
  }

  /** Get root node children. {@code == getNode().getChildren()} */
  public List<TraceNode> getChildren() {
    return getNode().getChildren();
  }

  /* For debug purposes */
  @Override
  public String toString() {
    return "TraceMessage "
        + getAspect()
        + ':'
        + getName()
        + " channel="
        + getChannel()
        + " attrs="
        + getAttrs();
  }
}
