/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import com.unitesk.aspectrace.provider.XmlFileParser;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Trace process manager. Controls adapters, sources and trace handlers.
 *
 * <p>Two main use scenarios:
 *
 * <p>1. PULL. For example, reading from already prepared XML-file using WSTX library which
 * represents pull-interface:
 *
 * <pre>TraceProcessorManager tpm = new TraceProcessorManager();
 * tpm.addProvider(pull_provider);// 1..* of providers
 * tpm.addListener(...);          // 1..* of listeners
 * tpm.process();
 * </pre>
 *
 * <p>2. PUSH. For example, converting data from a push-source:
 *
 * <pre>TraceProcessorManager tpm = new TraceProcessorManager();
 * tpm.addProvider(push_provider);// 1..* provider
 * tpm.addListener(...);          // 1..* listeners
 * push_provider.attachToManager(tpm);
 * <i>&lt;Processing messages from a push force;</i>
 * </pre>
 *
 * @author Sergey Groshev
 */
public class TraceProcessorManager {

  protected List<TraceProvider> providers = new ArrayList<TraceProvider>();

  private static final Map<String, String> defaultAdapterClazz =
      TraceUtils.createStringMap(
          "", "com.unitesk.aspectrace.common.CommonAdapter",
          "error", "com.unitesk.aspectrace.error.ErrorAdapter",
          "coverage", "com.unitesk.aspectrace.coverage.CoverageAdapter",
          "impl", "com.unitesk.aspectrace.impl.ImplAdapter",
          "mediator", "com.unitesk.aspectrace.media.MediatorAdapter",
          "oracle", "com.unitesk.aspectrace.oracle.OracleAdapter",
          "fsm", "com.unitesk.aspectrace.fsm.FsmAdapter",
          "test", "com.unitesk.aspectrace.test.TestAdapter",
          "requirement", "com.unitesk.aspectrace.requirement.RequirementAdapter");

  /** {@code aspect name +> adapter class name}. */
  protected Map<String, String> adapterClazz = new HashMap<String, String>(defaultAdapterClazz);

  /** {@code aspect name +> adapter}. */
  protected Map<String, AspectAdapter> activeAdapters = new HashMap<String, AspectAdapter>();

  // ------------------------------------------------------
  // Public interface

  /** Adds a trace source. */
  public synchronized void addProvider(TraceProvider provider) {
    providers.add(provider);
  }

  /** Adds a trace source from a XML-file. */
  public void addTrace(File trace) {
    addProvider(new XmlFileParser(trace));
  }

  /**
   * Sets an adapter class for future trace managers.
   *
   * @param aspect Aspect name.
   * @param clazzName Adapter class name.
   */
  public static synchronized void setAdapterClass(String aspect, String clazzName) {
    defaultAdapterClazz.put(aspect, clazzName);
  }

  /**
   * Sets an aspect adapter class for this trace manager.
   *
   * @param aspect Aspect name.
   * @param clazzName Class adapter name.
   */
  public void setLocalAdapterClass(String aspect, String clazzName) {
    adapterClazz.put(aspect, clazzName);
  }

  /**
   * Adds a listener.
   *
   * @throws Exception see documentation for {@link #getAdapter(String)} and for {@link
   *     AspectAdapter#addListener(AspectListener)}.
   */
  public final void addListener(String aspect, AspectListener listener) throws Exception {
    getAdapter(aspect).addListener(listener);
  }

  /**
   * Gets an aspect adapter. Adapter is created and is connected to handling.
   *
   * @throws IllegalArgumentException Unknown aspect or illegal adapter class.
   * @throws ClassCastException Illegal adapter class.
   * @throws IllegalStateException Can't change the configuration.
   * @throws Exception Create adapter instance error (see exceptions generated by Java Reflection
   *     API)
   */
  public final AspectAdapter getAdapter(String aspect) throws Exception {
    if (aspect == null) {
      throw new NullPointerException();
    }
    AspectAdapter adapter = activeAdapters.get(aspect);
    if (adapter == null) { // Try to create and attach adapter
      String className = adapterClazz.get(aspect);
      if (className == null) {
        throw new IllegalArgumentException("Unknown aspect: \"" + aspect + "\"");
      }
      Class<AspectAdapter> clazz = (Class<AspectAdapter>) Class.forName(className);
      adapter = clazz.getConstructor().newInstance();
      activeAdapters.put(aspect, adapter);
    }
    return adapter;
  }

  /**
   * Removes a listener.
   *
   * @throws IllegalStateException Can't change the configuration.
   */
  public final void removeListener(String aspect, AspectListener listener)
      throws IllegalStateException {
    if (aspect == null || listener == null) {
      throw new NullPointerException();
    }
    AspectAdapter adapter = activeAdapters.get(aspect);
    if (adapter == null) {
      return;
    }
    adapter.removeListener(listener);
  }

  /**
   * Begins handling in case if of working with active trace providers which work in the model PULL.
   * Method going through all providers one by one connecting adapters and calling {@link
   * TraceProvider#process process()}.
   *
   * <p>If provider works in the model PUSH then this method is never called.
   *
   * @throws IOException Input-Output error.
   * @throws IllegalStateException Handling has already begun.
   * @throws IllegalArgumentException Trace error.
   */
  public void process() throws IOException, IllegalStateException, IllegalArgumentException {
    for (TraceProvider provider : providers) {
      attachAdapters(provider);
      provider.process();
    }
  }

  /**
   * Attaches adapters and handlers to a provider. This method is called from {@link #process()} in
   * case of the model PULL or from {@link TraceProvider#attachToManager(TraceProcessorManager)} in
   * case of the model PUSH.
   *
   * @throws IllegalStateException Handling has already begun.
   * @throws IllegalArgumentException Provider is not connected to this trace manager.
   */
  protected void attachAdapters(TraceProvider provider)
      throws IllegalStateException, IllegalArgumentException {
    if (provider == null) {
      throw new NullPointerException();
    }
    if (!providers.contains(provider)) {
      throw new IllegalArgumentException("Provider not attached");
    }
    for (Map.Entry<String, AspectAdapter> e : activeAdapters.entrySet()) {
      provider.setAdapter(e.getKey(), e.getValue());
    }
  }
}
