/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import java.io.File;
import java.util.LinkedHashMap;
import java.util.List;

/**
 * Supplementary methods for trace processing and tracing.
 *
 * @author Sergey Groshev
 */
public class TraceUtils {

  /**
   * Converts a string array to mapping.
   *
   * @param info Array of the following type: [ key<sub>1</sub>, value<sub>1</sub>, key<sub>2</sub>,
   *     value<sub>2</sub>, ..., key<sub>n</sub>, value<sub>n</sub> ]
   * @return Mapping keys to the corresponding values. Mutable.
   */
  public static LinkedHashMap<String, String> createStringMap(String... info) {
    LinkedHashMap<String, String> lhm =
        new LinkedHashMap<String, String>(info.length / 2 + 1, 1.0f);
    for (int i = 0; i < info.length; i += 2) {
      lhm.put(info[i], info[i + 1]);
    }
    return lhm;
  }

  /**
   * Resolves an non-existing file name by a pattern.
   *
   * @param pattern File name pattern. <code>'*'</code> are replaced by numbers.
   * @return Non-existing file (Possibility of creating is not checked).
   */
  public static File resolveFilePattern(String pattern) {
    if (pattern.indexOf('*') < 0) {
      return new File(pattern);
    }
    for (int i = 0; ; i++) {
      File file = new File(pattern.replace("*", Integer.toString(i)));
      if (!file.exists()) {
        return file;
      }
    }
  }

  /**
   * Applies filters to trace messages.
   *
   * @param filters Filters.
   * @param msg Input message.
   * @return {@code null} is returned if a message must be ignored.
   */
  public static AbstractTraceMessage applyFilters(
      List<MessageStreamFilter> filters, AbstractTraceMessage msg) {
    if (filters == null || filters.isEmpty()) {
      return msg;
    }

    for (MessageStreamFilter f : filters) {
      msg = f.filter(msg);
      if (msg == null) {
        break;
      }
    }

    return msg;
  }
}
