/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.coverage;

import com.unitesk.aspectrace.LazyNode;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;
import com.unitesk.aspectrace.Tracer;
import com.unitesk.coverage.Coverage;
import com.unitesk.coverage.CoverageBase;
import com.unitesk.coverage.CoverageElement;
import com.unitesk.coverage.CoverageProcessor;
import com.unitesk.coverage.CoverageStructure;

/**
 * Tracer for the aspect "coverage".
 *
 * @author Sergey Groshev
 */
public class CoverageTracer {

  public static final String ASPECT = "coverage";

  /** Traces a coverage structure message. */
  public static void traceCoverageStructure(CoverageStructure cs) {
    TraceMessage msg = new TraceMessage(ASPECT, new CoverageStructureWrapper(cs));
    Tracer.traceMessage(msg);
  }

  /**
   * Traces a coverage structure message which structure is defined as a class with {@link Coverage
   * respectful annotation}.
   *
   * @param clazz Coverage class.
   * @throws IllegalArgumentException Class is not a coverage class or cannot be fully enumerated.
   */
  public static void traceCoverageStructure(Class<?> clazz) throws IllegalArgumentException {
    CoverageTracer.traceCoverageStructure(CoverageProcessor.toStructure(clazz));
  }

  /**
   * Traces a coverage element message.
   *
   * @param targetAspect Coverage structure aspect.
   * @param structName Coverage structure name.
   * @param elementChain Qualified name of coverage element. Must be not null array.
   */
  public static void traceCoverageElement(
      String targetAspect, String structName, String... elementChain) {
    if (elementChain == null || elementChain.length == 0) {
      elementChain = new String[] {""};
    }
    TraceNode tn =
        new TraceNode("element", "aspect", targetAspect, "cs", structName, "name", elementChain[0]);
    TraceNode curNode = tn;
    for (int i = 1; i < elementChain.length; i++) {
      TraceNode newNode = new TraceNode("element", "name", elementChain[i]);
      curNode.addChild(newNode);
      curNode = newNode;
    }

    traceNode(tn);
  }

  /** Traces a coverage element message. */
  public static void traceCoverageElement(CoverageElement ce) {
    CoverageStructure struct = ce.getStructure();
    if (struct == null || ce == struct.getRoot()) {
      throw new IllegalArgumentException("Unbound element");
    }
    traceCoverageElement(struct.getTargetAspect(), struct.getName(), ce.getNamePath());
  }

  /**
   * Traces a coverage element message. This method calls {@link
   * CoverageBase#getCEPath()}, thus element coverage object is not created and is not called for.
   *
   * @param cb Value corresponding to an element coverage tracing.
   * @throws IllegalArgumentException Error in a class structure or in data.
   */
  public static void traceCoverageElement(CoverageBase cb) throws IllegalArgumentException {
    traceCoverageElement(cb.getTargetAspect(), cb.getStructName(), cb.getCEPath());
  }

  /**
   * Traces a coverage element message.
   *
   * @param enumConst Enum coverage element corresponding to an element coverage tracing.
   * @throws IllegalArgumentException Const doesn't match an enum coverage.
   */
  public static void traceCoverageElement(Enum<?> enumConst) throws IllegalArgumentException {
    traceCoverageElement(
        CoverageProcessor.findElement(
            CoverageProcessor.toStructure(enumConst.getDeclaringClass()), enumConst.name()));
  }

  /**
   * Traces an impossible coverage element message.
   *
   * @param targetAspect Coverage structure aspect.
   * @param structName Coverage structure name.
   * @param elementChain Element name.
   */
  public static void traceElementImpossible(
      String targetAspect, String structName, String... elementChain) {
    TraceNode tn =
        new TraceNode(
            "impossible", "aspect", targetAspect, "cs", structName, "name", elementChain[0]);
    TraceNode curNode = tn;
    for (int i = 1; i < elementChain.length; i++) {
      TraceNode newNode = new TraceNode("element", "name", elementChain[i]);
      curNode.addChild(newNode);
      curNode = newNode;
    }

    traceNode(tn);
  }

  /** Traces an impossible coverage element message. */
  public static void traceElementImpossible(CoverageElement ce) {
    CoverageStructure struct = ce.getStructure();
    if (struct == null || ce == struct.getRoot()) {
      throw new IllegalArgumentException("Unbound element");
    }
    traceElementImpossible(struct.getTargetAspect(), struct.getName(), ce.getNamePath());
  }

  /** Traces an impossible coverage element message. */
  public static void traceElementImpossible(CoverageBase cb) {
    traceElementImpossible(cb.getTargetAspect(), cb.getStructName(), cb.getCEPath());
  }

  private static void traceNode(TraceNode tn) {
    TraceMessage msg = new TraceMessage(ASPECT, tn);
    Tracer.traceMessage(msg);
  }

  // ------------------------------------------------------

  /** LazyNode usage example. */
  protected static class CoverageStructureWrapper implements LazyNode {

    CoverageStructure cs;

    CoverageStructureWrapper(CoverageStructure cs) {
      this.cs = cs;
    }

    @Override
    public TraceNode toNode() {
      TraceNode tn =
          new TraceNode("structure", "aspect", cs.getTargetAspect(), "name", cs.getName());
      fillChildren(tn, cs.getElements());
      return tn;
    }
  }

  static void fillChildren(TraceNode tn, CoverageElement[] elements) {
    for (int i = 0; i < elements.length; i++) {
      CoverageElement[] children = elements[i].getChildren();
      TraceNode node =
          new TraceNode(
              "element",
              new String[] {
                "name", elements[i].getName()
              }); // Force usage of constructor with (...)
      if (elements[i].isForcedLeaf() && children.length != 0) {
        node.setAttr("isLeaf", "true");
      }
      tn.addChild(node);
      fillChildren(node, children);
    }
  }
}
