/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.fsm;

import com.unitesk.aspectrace.MessageLevel;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;
import com.unitesk.aspectrace.Tracer;
import com.unitesk.aspectrace.error.ErrorTracer;

import java.util.Map;

/**
 * Tracer for "fsm" aspect.
 * Used only for information about UniTESK scenario state machine manager for now.
 *
 * @author Sergey Groshev
 */
public class FsmTracer {

  public static final String ASPECT = "fsm";

  public static final int FAILURE_INDETERM = 0;
  private static final String[] shortDescs = {"Indeterministic behavior"};

  private static int fsmCounter = 0;
  private static int transCounter = 0;

  private static void traceNode(TraceNode tn) {
    TraceMessage msg = new TraceMessage(ASPECT, tn);
    Tracer.traceMessage(msg);
  }

  private static String createFsmId(int id) {
    return "dfsm." + id;
  }

  // ------------------------------------------------------

  /** Begins a "fsm" trace.
   *  @return SM id for future tracing.
   * */
  public static int traceFsmStart(String fullName) {
    int res;
    synchronized (FsmTracer.class) {
      res = ++fsmCounter;
    }
    String id = "dfsm." + res;
    Tracer.traceTagStart(ASPECT, id, null);

    TraceNode tn = new TraceNode("start", "name", id, "full-name", fullName);
    traceNode(tn);
    return res;
  }

  public static void traceFsmEnd(int fsmId) {
    String fid = createFsmId(fsmId);
    TraceNode tn = new TraceNode("end", "name", fid);
    traceNode(tn);

    Tracer.traceTagEnd(ASPECT, fid, null);
  }

  public static void traceState(int fsmId, int stateId, String stateName) {
    TraceNode tn =
        new TraceNode(
            "state", "id", Integer.toString(stateId), "name", stateName, "fsm", createFsmId(fsmId));
    traceNode(tn);
  }

  public static void traceTransitionStart(int fsmId, String transName) {
    String fid = createFsmId(fsmId);
    Tracer.traceTagStart(ASPECT, transName, fid);

    TraceNode tn = new TraceNode("start-transition", "name", transName, "fsm", fid);
    traceNode(tn);
  }

  /**
   * @return Transition ID for using in {@link #traceTransitionEndEx(int, String,
   *     int)}.
   */
  public static int traceTransitionStartEx(int fsmId, String transName) {
    int res;
    synchronized (FsmTracer.class) {
      res = ++transCounter;
    }
    String tid = transName + ":" + res;
    String fid = createFsmId(fsmId);

    Tracer.traceTagStart(ASPECT, transName, fid);
    Tracer.traceTagStart(ASPECT, tid, transName);
    TraceNode tn =
        new TraceNode(
            "start-transition", "name", transName, "fsm", fid, "id", Integer.toString(res));
    traceNode(tn);

    return res;
  }

  public static void traceTransitionEnd(int fsmId, String transName) {
    String fid = createFsmId(fsmId);
    TraceNode tn = new TraceNode("end-transition", "name", transName, "fsm", fid);
    traceNode(tn);

    Tracer.traceTagEnd(ASPECT, transName, fid);
  }

  public static void traceTransitionEndEx(int fsmId, String transName, int transId) {
    String tid = transName + ":" + transId;
    String fid = createFsmId(fsmId);

    TraceNode tn =
        new TraceNode(
            "end-transition", "name", transName, "fsm", fid, "id", Integer.toString(fsmId));
    traceNode(tn);

    Tracer.traceTagEnd(ASPECT, tid, transName);
    Tracer.traceTagEnd(ASPECT, transName, fid);
  }

  public static void traceIterationVariable(int fsmId, String type, String name, Object value) {
    TraceNode tn =
        new TraceNode(
            "iteration-variable",
            "type",
            type,
            "name",
            name,
            "value",
            String.valueOf(value),
            "fsm",
            createFsmId(fsmId));
    traceNode(tn);
  }

  public static void traceSummary(int fsmId, Map<String, String> properties) {
    TraceNode tn = new TraceNode("summary", "fsm", createFsmId(fsmId));
    for (Map.Entry<String, String> me : properties.entrySet()) {
      tn.addChild(new TraceNode("property", "name", me.getKey(), "value", me.getValue()));
    }
    traceNode(tn);
  }

  public static void traceBrowserMessage(int fsmId, MessageLevel level, String msg) {
    TraceNode tn =
        new TraceNode("browser-message", "fsm", createFsmId(fsmId), "level", level.toString());
    tn.addTextChild(msg);
    traceNode(tn);
  }

  /** Traces aspect {@code error} error message. */
  public static void traceFailure(int kind, String desc) {
    ErrorTracer.traceFailure(ASPECT, shortDescs[kind], desc, null, 2);
  }
}
