/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.html;

import com.unitesk.aspectrace.sampleApp.SummaryCollector;

import java.io.File;
import java.io.PrintWriter;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

public class HeaderContext extends HtmlReportPage {
  public static final int REPRESENTED_LIST = 5;
  ReportGenerator repGen;
  HtmlReportPage callingPage;

  public static int getRepresentedList() {
    return REPRESENTED_LIST;
  }

  public HeaderContext(ReportGenerator repGen, HtmlReportPage callingPage) {
    this.repGen = repGen;
    this.callingPage = callingPage;
  }

  public File reportDir;
  public String title = "Generated";
  public Map<String, String> summaryInfo = new HashMap<String, String>();
  public ArrayList<String> summaryKeys = new ArrayList<String>();
  private ArrayList<String> reportTypes = new ArrayList<String>();

  /**
   * Filter for displayed pages. May be overridden by descendants. Default realization:
   * all pages except this one which are not sections,
   * don't have index page and have not-null path.
   */
  public boolean accept(HtmlReportPage rp) {
    return rp != this
        && rp.getMasterPage() == null // Calls getParentPage().
        && rp.getPath() != null;
  }

  @Override
  public void generatePageBody(PrintWriter pw) {}

  public ArrayList<String> getPages() {
    ArrayList<String> result = new ArrayList<String>();
    for (HtmlReportPage p : repGen.getHtmlPages()) {
      if (!accept(p)) {
        continue;
      }

      HtmlReportPage re = (HtmlReportPage) p;
      result.add(re.getAspect() + re.generateRef(callingPage, null));
    }
    return result;
  }

  public ArrayList<String> getPagesFilter(String aspectFilter) {
    ArrayList<String> result = new ArrayList<String>();

    labelPages:
    for (HtmlReportPage p : repGen.getHtmlPages()) {
      boolean isFiltered = false;
      try {
        if (!accept(p)) {
          continue;
        }

        Class clazz = p.getClass();
        Method hasGetAspectMethod = clazz.getMethod("getAspect", null);
        if (hasGetAspectMethod != null) {
          String aspect = p.getAspect();
          if (aspect == null) {
            if (aspectFilter.endsWith("null")) {
              HtmlReportPage re = (HtmlReportPage) p;
              result.add(re.generateRef(callingPage, null));
            }
            continue labelPages;
          }

          isFiltered = aspect.equals(aspectFilter);
          if (isFiltered) {
            HtmlReportPage re = (HtmlReportPage) p;
            result.add(re.generateRef(callingPage, null));
          }
        }

      } catch (SecurityException se) {
        se.printStackTrace();
      } catch (NoSuchMethodException nsme) {
        continue;
      }
    }
    return result;
  }

  public ArrayList<String> getPagesFilterCutted(String aspectFilter) {
    ArrayList<String> result = new ArrayList<String>();
    ArrayList<String> listAll = getPagesFilter(aspectFilter);
    if (listAll.size() <= REPRESENTED_LIST) {
      result = listAll;
    } else {
      for (int i = 0; i < REPRESENTED_LIST; i++) {
        result.add(listAll.get(i));
      }
    }
    return result;
  }

  @Override
  public String getTitle() {
    return "Generated report";
  }

  public int getAspectCount(String aspectName) {
    int result = getPagesFilter(aspectName).size();

    return result;
  }

  public int getTracesCount() {
    return SummaryCollector.getKeys().size();
  }

  public ArrayList<String> getSummaryKeys() {
    return summaryKeys;
  }

  public int getMaxStates() {
    int result = 0;
    for (String key : SummaryCollector.getKeys()) {
      String statesString = SummaryCollector.getAttrValue(key, "States");
      String states = statesString.substring(0, statesString.indexOf("/"));
      int numberOfStates = Integer.parseInt(statesString.substring(0, statesString.indexOf("/")));
      if (numberOfStates > result) {
        result = numberOfStates;
      }
    }
    return result;
  }

  public String getMaxValue(String valueName) {
    int intResult = 0;
    String result = "unknown";
    for (String key : SummaryCollector.getKeys()) {
      String statesString = SummaryCollector.getAttrValue(key, valueName);
      int value = -1;
      int slashIndex = statesString.indexOf("/");
      if (slashIndex > 0) {
        value = Integer.parseInt(statesString.substring(0, slashIndex));
        if (value >= intResult) {
          intResult = value;
          result = "" + intResult;
        }
      }
    }
    return result;
  }

  public int getMaxTime() {
    int result = 0;
    for (String key : SummaryCollector.getKeys()) {
      String timeString = SummaryCollector.getAttrValue(key, "Traversal");
      int commaIndex = timeString.indexOf(",");
      int bracketIndex = timeString.indexOf(" (", timeString.indexOf(","));
      if ((bracketIndex > 0)) {
        String time = timeString.substring(commaIndex + 2, bracketIndex);
        int timeInt = Integer.parseInt(time);
        if (timeInt > result) {
          result = timeInt;
        }
      }
    }
    return result;
  }

  public void setSummaryKeys() {
    summaryKeys.add("TestRun");
    summaryKeys.add("Errors");
    summaryKeys.add("Coverage");
    summaryKeys.add("States");
    summaryKeys.add("Transitions");
    summaryKeys.add("Nondeterminism");
    summaryKeys.add("Time");
    summaryKeys.add("Computers number");
  }

  public String getSummaryInfoValue(String key) {
    return summaryInfo.get(key);
  }

  public Map<String, String> getSummaryInfo() {
    return summaryInfo;
  }

  public void setSummaryInfo(Map<String, String> summaryInfo) {
    this.summaryInfo = summaryInfo;
  }

  public void setReportTypes(ArrayList<String> reportTypes) {
    this.reportTypes = reportTypes;
  }

  public ArrayList<String> getReportTypes() {
    return reportTypes;
  }

  public String getReportType() {
    return "summary";
  }
}
