/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.html;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.LinkedHashSet;

/**
 * A page or a section of an HTML-report. The page is a single HTML-file, the section is
 * a selected part of other(parent) page.
 *
 * <p><b>Section:</b><br>
 *
 * <ul>
 *   <li>{@link #getParentPage()} returns not {@code null}
 * </ul>
 *
 * <p><b>Common page:</b><br>
 *
 * <ul>
 *   <li>{@link #getParentPage()} returns {@code null}
 *   <li>Has a not-{@code null} path
 *   (see {@link ReportPage#getPath()} and {@link ReportPage#setPath})
 *   <li>{@link #generatePageBody(PrintWriter)} generates a page base text and
 *   inserts child pages text in it if necessary.
 * </ul>
 *
 * <p><b>Stream page:</b><br>
 *
 * <ul>
 *   <li>Inherits class {@link Streamable}.
 *   <li>{@link Streamable#useStream()} returns {@code true}
 *   <li>{@link #getParentPage()} returns {@code null}
 *   <li>Has a not-{@code null} path
 *   (see {@link ReportPage#getPath()} and {@link ReportPage#setPath})
 *   <li>Method {@link Streamable#startBody} starts a page body generation;
 *   the base section of the page is generated by corresponding trace handlers;
 *   Method {@link Streamable#finishBody} finishes the generation.
 * </ul>
 */
public abstract class HtmlReportPage extends ReportPage {

  private String title;
  private String style;

  protected LinkedHashSet<PageRef> pageRefs = new LinkedHashSet<PageRef>();

  private HtmlReportPage parentPage = null;
  private String parentMark;
  private HtmlReportPage master;

  public HtmlReportPage() {}

  public HtmlReportPage(String path) {
    super(path);
  }

  // ------------------------------------------------------
  // Page hierarchy.

  /**
   * Gets a parent page if there is such. The sections' level nesting maybe more than 1.
   * The section generation is done by the parent page.
   *
   * @return {@code null} for a common page and a reference for a parent page for a section.
   */
  public HtmlReportPage getParentPage() {
    return parentPage;
  }

  /** Gets a current section's mark inside a parent page. */
  public String getParentMark() {
    return parentMark;
  }

  /**
   * Sets a parent page. If the parent page is set then this page becomes it's section and
   * stops being generated in filesystem by itself.
   * Requests for generator, path, coding, forming references for this page and maybe something else
   * is redirected for the parent page.
   *
   * @param parent Parent page or {@code null}, if this page must become independent.
   * @param parentMark Mark of this page inside parent page.
   */
  public void setParent(HtmlReportPage parent, String parentMark) {
    this.parentPage = parent;
    this.parentMark = parentMark;
  }

  /**
   * Sets a master page. Existence of the master page may be checked, for example,
   * during a list of report pages generation.
   *
   * @param master Master page.
   */
  public void setMasterPage(HtmlReportPage master) {
    this.master = master;
  }

  /**
   * Gets a master page if there is such.
   *
   * @return Master page. If this page is a section then a parent page is returned.
   */
  public HtmlReportPage getMasterPage() {
    HtmlReportPage parent = getParentPage();
    if (parent != null) {
      return parent;
    }
    return master;
  }

  // ---------- Redirections to parent page.

  /** {@inheritDoc} If parent page is defined then it's generator is called. */
  @Override
  public ReportGenerator getGenerator() {
    HtmlReportPage parent = getParentPage();
    if (parent != null) {
      return parent.getGenerator();
    } else {
      return super.getGenerator();
    }
  }

  /** {@inheritDoc} If parent page is defined then it's path is called. */
  @Override
  public String getPath() {
    HtmlReportPage parent = getParentPage();
    if (parent != null) {
      return parent.getPath();
    }
    return super.getPath();
  }

  /** {@inheritDoc} If parent page is defined then it's coding is called. */
  @Override
  public String getCharset() {
    HtmlReportPage parent = getParentPage();
    if (parent != null) {
      return parent.getCharset();
    } else {
      return super.getCharset();
    }
  }

  // ------------------------------------------------------
  // HTML-code generation.

  /**
   * {@inheritDoc}<br>
   * Input-Output channel creation is assigned to a generator.<br>
   * Pre: {@link #getParentPage()} == null
   */
  @Override
  public void generate() throws IOException {
    PrintWriter pw = getGenerator().createWriter(this);
    generatePage(pw);
    pw.close();
  }

  /**
   * Generates a page full code:<br>
   * Head<br>
   * Body<br>
   * RefSection<br>
   * Tail<br>
   *
   * <p>If the page is included in another then only the "body" is generated.
   */
  public void generatePage(PrintWriter pw) {
    if (getParentPage() != null) {
      generatePageBody(pw);
      return;
    }
    generatePageHead(pw);
    generatePageBody(pw);
    generateRefSection(pw);
    generatePageTail(pw);
  }

  /** Generates page head. */
  protected void generatePageHead(PrintWriter pw) {
    pw.println("<html>");
    pw.println("<head>");
    generateHeadContents(pw);
    pw.println("</head>");
    pw.println("<body>");
  }

  /** Generates content of tag {@code <head>}. */
  protected void generateHeadContents(PrintWriter pw) {
    pw.println(
        "  <meta http-equiv=\"Content-Type\" content=\"text/html; charset=" + getCharset() + "\">");
    pw.println("  <title>" + encodeSimple(getTitle()) + "</title>");
    String css = getStyle();
    if (css != null) {
      pw.println(
          "  <style type=\"text/css\" media=\"all\">@import \""
              + getRelativePathTo(css)
              + "\";</style>");
    }
  }

  /** Generates page tail. */
  protected void generatePageTail(PrintWriter pw) {
    pw.println("</body></html>");
  }

  /** Generates page body. */
  public abstract void generatePageBody(PrintWriter pw);

  // ------------------------------------------------------
  // Settings.

  /** Sets page title. */
  public void setTitle(String title) {
    this.title = title;
  }

  /**
   * Gets a page title. If the page title is not set explicitly then the default realization
   * returns path to this page.
   */
  public String getTitle() {
    if (title != null) {
      return title;
    } else {
      return getPath();
    }
  }

  /**
   * Sets CSS style.
   *
   * @param cssStyle Path relative to report root.
   */
  public void setStyle(String cssStyle) {
    this.style = cssStyle;
  }

  /** Gets CSS style. */
  public String getStyle() {
    return style;
  }

  // ------------------------------------------------------
  // References.

  /**
   * Adds a reference to other page.
   *
   * @param otherPage Page.
   * @param mark Mark.
   * @return true if there has been none such reference.
   */
  public boolean addPageRef(HtmlReportPage otherPage, String mark) {
    return pageRefs.add(new PageRef(this, otherPage, mark));
  }

  /**
   * Adds a reference to other page.
   *
   * @param ref Reference. Must be from this page.
   * @return true if there has been none such reference.
   * @throws IllegalArgumentException Reference is not from this page.
   */
  public boolean addPageRef(PageRef ref) {
    if (ref.from != this) {
      throw new IllegalArgumentException("Illegal reference origin");
    }
    return pageRefs.add(ref);
  }

  /**
   * Gets a relative reference to a page. If this page is included in another then
   * generation is delegated to a parent page.
   *
   * @param from Page from where reference is.
   * @param mark Section's anchor. Null for reference to whole page.
   * @return Reference type <i>path/.../path/file#section</i>
   */
  public String getRelativeRef(HtmlReportPage from, String mark) {
    HtmlReportPage parent = getParentPage();
    if (parent != null) { // Redirect to parent page.
      return parent.getRelativeRef(from, getParentMark());
    }

    if (mark == null) {
      return getRelativePathFrom(from);
    } else {
      return getRelativePathFrom(from) + '#' + mark;
    }
  }

  /**
   * Generates a reference to the section <code>mark</code> of this page from the page <code>from</code>.
   * If this page is included in another then generation is delegated to parent page.
   *
   * @param from Page from where reference is.
   * @param mark Section's anchor. Null for reference to whole page.
   * @return HTML-code of reference.
   */
  public String generateRef(HtmlReportPage from, String mark) {
    HtmlReportPage parent = getParentPage();
    if (parent != null) { // Redirect to parent page.
      return parent.generateRef(from, getParentMark());
    }

    String markPostfix = mark == null ? "" : "#" + mark;
    return "<a href=\""
        + getRelativeRef(from, mark)
        + "\">"
        + encodeSimple(getTitle())
        + markPostfix
        + "</a>";
  }

  /** Simple encoder of HTML special-symbols. */
  protected static String encodeSimple(String str) {
    if (str == null) {
      return "";
    }
    int len = str.length();
    StringBuilder sb = new StringBuilder(len * 2 + 4);
    for (int i = 0; i < len; i++) {
      char ch = str.charAt(i);
      switch (ch) {
        case '<':
          sb.append("&lt;");
          break;
        case '>':
          sb.append("&gt;");
          break;
        case '&':
          sb.append("&amp;");
          break;
        default:
          sb.append(ch);
      }
    }
    if (len == sb.length()) {
      return str;
    } else {
      return sb.toString();
    }
  }

  /**
   * Generates a simple text reference for the section <code>mark</code> of this page from the page
   * <code>from</code>.
   * If this page is included in another then generation is delegated to parent page.
   *
   * @param from Page from where reference is.
   * @param mark Section's anchor. Null for reference to whole page.
   * @param refText Text of reference.
   * @return HTML-code of reference.
   */
  public String generateTextRef(HtmlReportPage from, String mark, String refText) {
    return "<a href=\"" + getRelativeRef(from, mark) + "\">" + refText + "</a>";
  }

  /** Generates reference section. */
  protected void generateRefSection(PrintWriter pw) {
    if (pageRefs.size() == 0) {
      return;
    }
    pw.println("<br/><hr/><h2>Context links</h2>");
    generateRefs(pw, false);
  }

  /** Generates block of references from this page without heads. */
  public void generateRefs(PrintWriter pw, boolean compact) {
    String delim = compact ? "&nbsp;" : "<br/>\n";
    for (PageRef pgRef : pageRefs) {
      pw.print(pgRef.generateRef());
      pw.print(delim);
    }
  }

  // ------------------------------------------------------
  // Streamable pages support.

  /**
   * A streamable page. If the method {@link #useStream()} returns {@code false} then
   * the page behaves like a normal page.<br>
   * Otherwise the method {@link #generatePageBody(PrintWriter)} is never called. The method {@link
   * #startBody(PrintWriter)} prepares a stream generation of a main page body. The method
   * {@link #finishBody(PrintWriter)} finishes all necessary generation of the body page.
   */
  public abstract static class Streamable extends HtmlReportPage {

    protected PrintWriter pw;

    public Streamable() {}

    public Streamable(String path) {
      super(path);
    }

    /** If method returns false then page behaves like a normal page. */
    public boolean useStream() {
      return getParentPage() == null;
    }

    /**
     * Generates a page body. The method would be called only if <code>
     * {@link #useStream()} == false</code> - in this case it needs to be overridden.
     */
    @Override
    public void generatePageBody(PrintWriter pw) {}

    /**
     * Saves PrintWriter and starts page generation.
     *
     * <p>Standard scheme:<br>
     * <i> Head<br>
     * startBody()</i>
     */
    public void startGeneration(PrintWriter pw) {
      this.pw = pw;
      generatePageHead(pw);
      startBody(pw);
    }

    /**
     * Ends page generation.
     *
     * <p>Standard scheme:<br>
     * <i> finishBody()<br>
     * RefSection<br>
     * Tail</i>
     */
    public void finishGeneration(PrintWriter pw) {
      finishBody(pw);
      generateRefSection(pw);
      generatePageTail(pw);
    }

    /** Starts body generation. */
    public void startBody(PrintWriter pw) {}

    /** Ends body generation. */
    public void finishBody(PrintWriter pw) {}
  }
}
