/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.html;

import com.unitesk.aspectrace.ref.ReferenceManager;
import com.unitesk.aspectrace.ref.ReferenceResolver;
import com.unitesk.aspectrace.requirement.AllDBReports;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * HTML-report generator. Supports common generation and stream generation, reference manager and
 * auto-placing of context references.
 *
 * @author Sergey Groshev
 */
public class ReportGenerator {

  protected ReferenceManager<HtmlReportPage> refMan = new ReferenceManager<HtmlReportPage>();
  protected Map<List<String>, Boolean> autoRefs = new HashMap<List<String>, Boolean>();

  protected LinkedHashSet<ReportPage> pages = new LinkedHashSet<ReportPage>();

  private File dir;
  private String defaultCharset = "utf-8";

  protected IdentityHashMap<HtmlReportPage.Streamable, PrintWriter> streams =
      new IdentityHashMap<HtmlReportPage.Streamable, PrintWriter>();

  /** Sets generation directory. Directory would be created if necessary. */
  public void setGenerationDir(File dir) {
    this.dir = dir;
    if (!dir.exists()) {
      dir.mkdirs();
    }
  }

  public File getGenerationDir(File dir) {
    return dir;
  }

  /** Gets generation directory. */
  public File getGenerationDir() {
    return dir;
  }

  /** Gets default coding for pages. */
  public String getDefaultCharset() {
    return defaultCharset;
  }

  /** Sets default coding for pages. */
  public void setDefaultCharset(String charset) {
    defaultCharset = charset;
  }

  // ------------------------------------------------------
  // Pages.

  /**
   * Adds a page in a report. If the page is {@link HtmlReportPage.Streamable streamble
   * HTML-page} then generation begins immediately.
   */
  public void addPage(ReportPage rp) {
    if (!pages.add(rp)) {
      return;
    }

    rp.setGenerator(this);

    if (rp instanceof HtmlReportPage.Streamable) {
      HtmlReportPage.Streamable strmbl = (HtmlReportPage.Streamable) rp;
      if (strmbl.useStream()) { // Should also check parent
        try {
          startGeneration(strmbl); // Will check path
        } catch (IOException e) {
          throw new RuntimeException(e);
        }
      }
    }
  }

  /**
   * Adds pages in report.
   *
   * @see ReportGenerator#addPage(ReportPage)
   */
  public void addPages(ReportPage[] ps) {
    for (int i = 0; i < ps.length; i++) {
      addPage(ps[i]);
    }
  }

  /**
   * Adds a common index page.
   *
   * @return Added page.
   */
  public HtmlReportPage addIndexPage() {
    HtmlReportPage hrp = new TemplateIndexPage();
    addPage(hrp);
    return hrp;
  }

  public HtmlReportPage addAllDtBsPage() {
    HtmlReportPage hrp = new AllDBReports();
    addPage(hrp);
    return hrp;
  }

  public HtmlReportPage addDetailsPage(HtmlReportPage ti) {
    HtmlReportPage hrp = new TemplateDetailsPage(ti);
    addPage(hrp);
    return hrp;
  }

  public HtmlReportPage addGenericReport(
      String name, String aspect, HtmlReportPage ti, String reportName) {
    HtmlReportPage hrp = new TemplateGenericReportsPage(name, aspect, ti, reportName);
    addPage(hrp);
    return hrp;
  }

  /**
   * Adds a resource.
   *
   * @param inPath Source path in filesystem or in classpath.
   * @param outPath Result path relative to generation root.
   * @return Added resource object.
   */
  public StaticResource addResource(String inPath, String outPath) {
    StaticResource sr = new StaticResource(inPath, outPath);
    addPage(sr);
    return sr;
  }

  /** Gets all pages and report sections. */
  public ReportPage[] getPages() {
    return pages.toArray(new ReportPage[pages.size()]);
  }

  /** Gets all HTML-pages and reports sections. */
  public HtmlReportPage[] getHtmlPages() {
    List<HtmlReportPage> list = new ArrayList<HtmlReportPage>();
    for (ReportPage rp : pages) {
      if (rp instanceof HtmlReportPage) {
        list.add((HtmlReportPage) rp);
      }
    }
    return list.toArray(new HtmlReportPage[list.size()]);
  }

  // ------------------------------------------------------
  // Cross-references.

  private static List<String> pair(String fromAspect, String toAspect) {
    List<String> list = new ArrayList<String>(2);
    list.add(fromAspect);
    list.add(toAspect);
    return list;
  }

  /**
   * Adds auto referencing (one-sided) between HTML-pages via resolvers and tags.
   */
  public void addAutoRef(String fromAspect, String toAspect) {
    addAutoRef(fromAspect, toAspect, false);
  }

  /**
   * Adds auto referencing (one-sided or two-sided) between HTML-pages of two aspects.
   */
  public void addAutoRef(String fromAspect, String toAspect, boolean bidirect) {
    autoRefs.put(pair(fromAspect, toAspect), Boolean.valueOf(bidirect));
  }

  /** Removes auto referencing between HTML-pages of two aspects. */
  public void removeAutoRef(String fromAspect, String toAspect) {
    autoRefs.remove(pair(fromAspect, toAspect));
  }

  /** Adds resolver for auto referencing. */
  public void addRefResolver(String aspect, ReferenceResolver<? extends HtmlReportPage> resolver) {
    refMan.addResolver(aspect, (ReferenceResolver<HtmlReportPage>) resolver);
  }

  /** Removes resolver from auto referencing. */
  public void removeRefResolver(
      String aspect, ReferenceResolver<? extends HtmlReportPage> resolver) {
    refMan.removeResolver(aspect, (ReferenceResolver<HtmlReportPage>) resolver);
  }

  /** Resolves text reference into HTML-page. */
  public HtmlReportPage resolve(String aspect, String channel, String ref) {
    return refMan.resolve(aspect, channel, ref);
  }

  /**
   * Resolves a text reference into multiple HTML-pages.
   * Here a list can be replaced by an array cause dynamic type is known.
   */
  public HtmlReportPage[] resolveMulti(String aspect, String channel, String ref) {
    List<HtmlReportPage> results = refMan.resolveMulti(aspect, channel, ref);
    return results.toArray(new HtmlReportPage[results.size()]);
  }

  protected void buildAutoRefs() {
    for (ReportPage rp : pages) {
      if (!(rp instanceof HtmlReportPage)) {
        continue;
      }
      HtmlReportPage hrp = (HtmlReportPage) rp;
      String asp = hrp.getAspect();
      if (asp == null) {
        continue;
      }
      for (Map.Entry<List<String>, Boolean> ar : autoRefs.entrySet()) {
        String fromAspect = ar.getKey().get(0);
        if (!asp.equals(fromAspect)) {
          continue;
        }
        String toAspect = ar.getKey().get(1);
        Set<String> marks = hrp.getTagMarks().get(toAspect);
        if (marks == null) {
          continue;
        }
        boolean bidirect = ar.getValue().booleanValue();
        for (String mark : marks) {
          HtmlReportPage[] targets = resolveMulti(toAspect, hrp.getAssociatedChannel(), mark);
          for (int i = 0; i < targets.length; i++) {
            hrp.addPageRef(targets[i], mark);
            if (bidirect) {
              targets[i].addPageRef(hrp, null);
            }
          }
        }
      }
    }
  }

  // ------------------------------------------------------
  // Generation.

  /**
   * Generates all report pages except:
   *
   * <ul>
   *   <li>HTML-pages included as sections in other HTML-pages;
   *   <li>Pages without path;
   *   <li>Pages which are generating on fly now. Generation ends for them.
   * </ul>
   */
  public void generateReport() throws IOException, IllegalStateException {
    buildAutoRefs();
    for (ReportPage p : pages) {
      p.buildRefs();
    }
    for (ReportPage p : pages) {
      if (!p.getClass().toString().contains("CoverageCollectorCTESK")) {
        continue;
      }
      if (p.getPath() == null) {
        continue;
      }
      if (p instanceof HtmlReportPage) {
        HtmlReportPage hp = (HtmlReportPage) p;
        if (hp.getParentPage() != null) {
          continue;
        }
      }
      if (p instanceof HtmlReportPage.Streamable) {
        HtmlReportPage.Streamable strmbl = (HtmlReportPage.Streamable) p;
        if (strmbl.useStream()) {
          finishGeneration(strmbl);
          continue;
        }
      }
      p.generate();
    }
    for (ReportPage p : pages) {
      if (p.getClass().toString().contains("CoverageCollectorCTESK")) {
        continue;
      }
      if (p.getPath() == null) {
        continue;
      }
      if (p instanceof HtmlReportPage) {
        HtmlReportPage hrp = (HtmlReportPage) p;
        if (hrp.getParentPage() != null) {
          continue;
        }
      }

      if (p instanceof HtmlReportPage.Streamable) {
        HtmlReportPage.Streamable strmbl = (HtmlReportPage.Streamable) p;
        if (strmbl.useStream()) {
          finishGeneration(strmbl);
          continue;
        }
      }

      p.generate();
    }
  }

  protected void startGeneration(HtmlReportPage.Streamable strmbl) throws IOException {
    if (strmbl.getPath() == null) {
      return;
    }
    PrintWriter pw = createWriter(strmbl);
    streams.put(strmbl, pw);
    strmbl.startGeneration(pw);
  }

  protected void finishGeneration(HtmlReportPage.Streamable strmbl) {
    PrintWriter pw = streams.get(strmbl);
    if (pw == null) {
      return;
    }
    strmbl.finishGeneration(pw);
    pw.close();
    streams.remove(strmbl);
  }

  protected PrintWriter createWriter(HtmlReportPage hp) throws IOException {
    File fl = new File(getGenerationDir(), hp.getPath());
    File flDir = fl.getParentFile();
    if (flDir != null && !flDir.isDirectory()) {
      flDir.mkdirs(); // Ignore errors here
    }
    return new PrintWriter(fl, hp.getCharset());
  }
}
