/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.oracle;

import com.unitesk.aspectrace.coverage.CoverageTracer;
import com.unitesk.coverage.CoverageElement;
import com.unitesk.coverage.CoverageStructure;

import java.util.ArrayList;
import java.util.List;

/**
 * A tracer of oracle marked paths.
 *
 * <p>Использование:<br>
 * <code>OracleTracer.traceOracleStart(...);<br>
 * <b>MarkedPath path = new MarkedPath(oracleSignature)</b>;<br>
 * OracleTracer.traceFormula(...);<br>
 * <b>path.traceMark(mark1);</b><br>
 * OracleTracer.traceFormula(...);<br>
 * <b>path.traceMark(mark2);</b><br>
 * OracleTracer.traceFormula(...);<br>
 * <b>path.traceBranch(branch);</b><br>
 * OracleTracer.traceMark(postMark1);<br>
 * OracleTracer.traceOracleEnd(...);<br>
 * </code>
 *
 * @see OracleTracer#traceBranch(String)
 * @see OracleTracer#traceMark(String)
 * @see CoverageTracer#traceCoverageElement(CoverageElement)
 * @author Sergey Groshev
 */
public class MarkedPath {

  // TODO? Deferred tracing. Useful for coverage element checks.
  // TODO? Post-marks.

  /** Prefix for coverage structure name. */
  public static final String CS_PREFIX = "marked path:";

  String signature;
  CoverageStructure cs;
  List<String> marks = new ArrayList<String>();

  public MarkedPath(String signature) {
    this.signature = signature;
  }

  public MarkedPath(CoverageStructure cs) {
    // TODO? Check whether coverage struct is really "marked path" type
    if (!cs.getTargetAspect().equals(OracleTracer.ASPECT)) {
      throw new IllegalArgumentException("Oracle coverage structure required");
    }
    this.cs = cs;
  }

  /** Trace and remember mark. */
  public void traceMark(String mark) {
    OracleTracer.traceMark(mark);
    marks.add(mark);
  }

  /** Trace branch, then trace collected marked path as CoverageElement. */
  public void traceBranch(String branch) {
    OracleTracer.traceBranch(branch);

    int count = marks.size() + 1;
    String[] chain = new String[count];
    chain[0] = "branch " + branch;
    for (int i = 1; i < count; i++) {
      chain[i] = marks.get(i - 1);
    }

    if (cs != null) {
      CoverageElement ce = cs.findElement(chain);
      if (ce == null) {
        throw new IllegalArgumentException("Marked path not in coverage structure: " + marks);
      }
      CoverageTracer.traceCoverageElement(ce);
    } else {
      CoverageTracer.traceCoverageElement(OracleTracer.ASPECT, CS_PREFIX + signature, chain);
    }
  }
}
