/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.oracle;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.Parameter;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;

import java.util.List;

/**
 * Adapter for aspect "oracle".
 *
 * @author Sergey Groshev
 */
public class OracleAdapter extends AspectAdapter {

  @Override
  public String getAspect() {
    return "oracle";
  }

  @Override
  public Class<OracleListener> getListenerClass() {
    return OracleListener.class;
  }

  @Override
  protected void processMessage(TraceMessage msg) throws IllegalArgumentException {
    super.processMessage(msg);

    String msgName = msg.getName();
    if (msgName.equals("start")) {
      OracleStartMessage osm = new OracleStartMessage(this);
      osm.signature = msg.getAttr("name");
      String id = msg.getAttr("id");
      if (id != null) {
        try {
          osm.id = Integer.parseInt(id);
        } catch (NumberFormatException nfe) {
          throw new IllegalArgumentException("Invalid start id:" + id, nfe);
        }
      }
      for (TraceNode tn : msg.getNode().getChildren()) {
        osm.parameters.add(Parameter.fromNode(tn));
      }
      for (OracleListener lst : getListeners()) {
        lst.processOracleStart(osm);
      }
    } else if (msgName.equals("end")) {
      OracleEndMessage oem = new OracleEndMessage(this);
      oem.signature = msg.getAttr("name");
      String id = msg.getAttr("id");
      if (id != null) {
        try {
          oem.id = Integer.parseInt(id);
        } catch (NumberFormatException nfe) {
          throw new IllegalArgumentException("Invalid end id:" + id, nfe);
        }
      }
      for (OracleListener lst : getListeners()) {
        lst.processOracleEnd(oem);
      }
    } else if (msgName.equals("pre")) {
      boolean verdict = Boolean.parseBoolean(msg.getAttr("verdict"));
      for (OracleListener lst : getListeners()) {
        lst.processPreEnd(verdict);
      }
    } else if (msgName.equals("post")) {
      boolean verdict = Boolean.parseBoolean(msg.getAttr("verdict"));
      for (OracleListener lst : getListeners()) {
        lst.processPostEnd(verdict);
      }
    } else if (msgName.equals("formula")) {
      FormulaMessage frmlMsg = new FormulaMessage(this, true);
      frmlMsg.id = Integer.parseInt(msg.getAttr("id"));
      frmlMsg.value = Boolean.parseBoolean(msg.getAttr("value"));
      for (OracleListener lst : getListeners()) {
        lst.processFormulaMessage(frmlMsg);
      }
    } else if (msgName.equals("formula-meaning")) {
      FormulaMessage frmlMsg = new FormulaMessage(this, false);
      frmlMsg.id = Integer.parseInt(msg.getAttr("id"));
      frmlMsg.formula = msg.getAttr("formula");
      for (OracleListener lst : getListeners()) {
        lst.processFormulaMessage(frmlMsg);
      }
    } else if (msgName.equals("mark") || msgName.equals("branch")) {
      MarkMessage mrkMsg = new MarkMessage(this);
      mrkMsg.kind = msgName.equals("mark") ? MarkMessage.MARK : MarkMessage.BRANCH;
      mrkMsg.name = msg.getAttr("name");
      for (OracleListener lst : getListeners()) {
        lst.processMark(mrkMsg);
      }
    } else {
      throw new IllegalArgumentException("Unsupported oracle node: " + msgName);
    }
  }

  @Override
  protected List<OracleListener> getListeners() {
    return (List<OracleListener>) super.getListeners();
  }
}
