/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.oracle;

import com.unitesk.aspectrace.Parameter;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;
import com.unitesk.aspectrace.Tracer;
import com.unitesk.aspectrace.error.ErrorTracer;

/**
 * A tracer for an aspect "oracle". Note: mark, branch and marked paths from them must be traced
 * via class {@link MarkedPath}
 *
 * @author Sergey Groshev
 */
public class OracleTracer {

  public static final String ASPECT = "oracle";

  public static final int FAILURE_PRE_VIOLATION = 0;
  public static final int FAILURE_POST_VIOLATION = 1;
  private static final String[] shortDescs = {"Precondition violation", "Postcondition violation"};

  private static int callCounter = 0;

  private static void traceNode(TraceNode tn) {
    TraceMessage msg = new TraceMessage(ASPECT, tn);
    Tracer.traceMessage(msg);
  }

  private static String idToTag(String signature, int callId) {
    return signature + ':' + callId;
  }

  public static void traceOracleStart(String signature, Parameter... params) {
    Tracer.traceTagStart(ASPECT, signature, null);

    TraceNode tn = new TraceNode("start", "name", signature);
    for (Parameter p : params) {
      tn.addChild(p.toNode());
    }

    traceNode(tn);
  }

  /**.
   *  @return callId for use in {@link #traceOracleEndEx(String, int)}. */
  public static int traceOracleStartEx(String signature, Parameter... params) {
    int res;
    synchronized (OracleTracer.class) {
      res = ++callCounter;
    }
    String callId = idToTag(signature, res);

    Tracer.traceTagStart(ASPECT, signature, null);
    Tracer.traceTagStart(ASPECT, callId, signature);

    TraceNode tn = new TraceNode("start", "name", signature, "id", Integer.toString(res));
    for (Parameter p : params) {
      tn.addChild(p.toNode());
    }
    traceNode(tn);

    return res;
  }

  public static void traceOracleEnd(String signature) {
    TraceNode tn = new TraceNode("end", "name", signature);
    traceNode(tn);
    Tracer.traceTagEnd(ASPECT, signature, null);
  }

  public static void traceOracleEndEx(String signature, int callId) {
    TraceNode tn = new TraceNode("end", "name", signature, "id", Integer.toString(callId));
    traceNode(tn);

    Tracer.traceTagEnd(ASPECT, idToTag(signature, callId), signature);
    Tracer.traceTagEnd(ASPECT, signature, null);
  }

  public static void tracePreVerdict(boolean verdict) {
    TraceNode tn = new TraceNode("pre", "verdict", Boolean.toString(verdict));
    traceNode(tn);
  }

  public static void tracePostVerdict(boolean verdict) {
    TraceNode tn = new TraceNode("post", "verdict", Boolean.toString(verdict));
    traceNode(tn);
  }

  public static void traceFormulaMeaning(int id, String formula) {
    TraceNode tn = new TraceNode("formula-meaning", "id", Integer.toString(id), "formula", formula);
    traceNode(tn);
  }

  public static void tracePrimeFormula(int id, boolean value) {
    TraceNode tn =
        new TraceNode("formula", "id", Integer.toString(id), "value", Boolean.toString(value));
    traceNode(tn);
  }

  public static void traceMark(String mark) {
    TraceNode tn = new TraceNode("mark", "name", mark);
    traceNode(tn);
  }

  public static void traceBranch(String branch) {
    TraceNode tn = new TraceNode("branch", "name", branch);
    traceNode(tn);
  }

  public static void traceFailure(int kind, String desc) {
    ErrorTracer.traceFailure(ASPECT, shortDescs[kind], desc, null, 2);
  }
}
