/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.provider;

import com.unitesk.aspectrace.tracer.XmlFormatter;

import java.io.IOException;
import java.io.InputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashMap;

/**
 * Провайдер трассы, переданной по сети в формате XML (такие трассы генерируются с помощью {@link
 * XmlFormatter} или приходят из CTESK).
 *
 * <p>Слушает TCP-порт с указанным номером. При вызове метода {@link XmlStreamParser#process()
 * process()} принимает <u>ровно 1</u> входящее соединение, читает весь переданный по нему поток
 * XML-сообщений и обрабатывает полученную трассу. При необходимости можно вызывать метод {@code
 * process()} несколько раз, при этом каждый раз будет обрабатываться очередное сетевое соединение.
 *
 * <p>Также используется классом {@link XmlSocketServer} для обработки принятых им соединений.
 *
 * @see XmlSocketServer
 * @author Sergey Groshev
 */
public class XmlSocketParser extends XmlStreamParser {

  /** Для работы в режиме сервера. */
  protected ServerSocket serverSock;

  protected int port;

  /** Для работы в режиме обработки одного соединения. */
  protected Socket connection;

  protected String channelPrefix;
  protected HashMap<String, String> channelIds = new HashMap<String, String>();

  private String traceDesc;

  /**
   * Конструктор.
   * @param port TCP-порт для приёма соединений.
   */
  public XmlSocketParser(int port) throws IOException {
    this.port = port;
    setTraceDesc("tcp:" + port);
    serverSock = new ServerSocket(port);
  }

  protected XmlSocketParser(Socket sock) {
    this.connection = sock;
  }

  /** Закрыть серверный сокет. */
  public synchronized void closeSocket() {
    try {
      if (serverSock != null) {
        serverSock.close();
        serverSock = null;
      }
      if (connection != null) {
        connection.close();
        connection = null;
      }
    } catch (IOException ioe) {
      // Silently ignore
    }
  }

  @Override
  protected void finalize() throws Throwable {
    try {
      closeSocket();
    } finally {
      super.finalize();
    }
  }

  protected void setChannelPrefix(String prefix) {
    channelPrefix = prefix;
    channelIds.clear();
  }

  protected String getChannelPrefix() {
    return channelPrefix;
  }

  // ------------------------------------------------------
  // XmlStreamParser

  @Override
  protected InputStream getInputStream() throws IOException, IllegalStateException {
    Socket sckt;
    if (serverSock != null) {
      sckt = serverSock.accept();
      System.err.println("XmlSocketParser: new connection accepted on port " + port);
    } else if (connection != null) {
      sckt = connection;
    } else {
      throw new IllegalStateException("No open sockets");
    }
    return sckt.getInputStream();
  }

  @Override
  protected String getSysId() {
    return null;
  }

  @Override
  protected String getTraceDesc() {
    return traceDesc;
  }

  public void setTraceDesc(String traceDesc) {
    this.traceDesc = traceDesc;
  }

  @Override
  protected String channelFor(String traceChannelId) {
    if (channelPrefix == null) {
      return traceChannelId;
    }
    String verbose = channelIds.get(traceChannelId);
    if (verbose == null) {
      verbose = channelPrefix + ':' + traceChannelId;
      channelIds.put(traceChannelId, verbose);
    }
    return verbose;
  }
}
