/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.provider;

import com.unitesk.aspectrace.AbstractTraceMessage;
import com.unitesk.aspectrace.DebugMessage;
import com.unitesk.aspectrace.TagMessage;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceProvider;
import com.unitesk.aspectrace.Tracer;
import com.unitesk.aspectrace.tracer.XmlFormatter;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;

/**
 * Сетевой сервер, принимающий XML-трассы из множества источников на удалённых машинах. Поддерживает
 * один порт с произвольным количеством одновременно установленных соединений. Смешивает в единый
 * поток сообщения изо всех входных соединений по мере их поступления из сети и синтаксического
 * разбора, при этом гарантируется относительный порядок сообщений только внутри одного канала.
 *
 * <p>Самостоятельно присваивает входным трассам уникальные идентификаторы каналов, но не извещает
 * об этом источники трассы. <b>(!!! возможны проблемы с сообщениями, содержащими в своём теле
 * идентификаторы каналов !!!)</b>
 *
 * <p>Работает в модели PULL. Обработка входящих соединений начинается в момент вызова метода {@link
 * #process()} и прекращается тогда, когда выполнены ВСЕ следующие условия:
 *
 * <ul>
 *   <li>все старые соединения закрыты
 *   <li>нет новых входящих соединений
 *   <li>ранее был вызван метод {@link #mayTerminate()} (метод также автоматически вызывается при
 *       закрытии любого соединения, если был установлен {@link #setAutoTerminate(boolean) флаг
 *       автозавершения}).
 * </ul>
 * После завершения обработки серверный сокет автоматически закрывается, и дальнейшая обработка
 * соединений этим экземпляром сервера уже невозможна.
 *
 * @see XmlSocketParser
 * @author Sergey Groshev
 */

// TODO: извещать удалённые источники трассы о присвоенных им идентификаторах каналов

public class XmlSocketServer extends TraceProvider {

  protected ServerSocket serverSock;
  protected int port;

  protected boolean shutdownFlag = false;
  protected boolean autoTerminate;

  protected PrintWriter messageLog;
  protected XmlFormatter logFormatter;

  /**
   * Конструктор.
   *
   * @param port TCP-порт для приёма соединений.
   * @param autoTerminate Параметр для {@link #setAutoTerminate(boolean)}
   * @throws IOException Ошибка ввода-вывода.
   */
  public XmlSocketServer(int port, boolean autoTerminate) throws IOException {
    this.port = port;
    serverSock = new ServerSocket(port);
    serverSock.setSoTimeout(1);
    setAutoTerminate(autoTerminate);
  }

  /**
   * Установить режим автоматического завершения.
   *
   * @param autoTerminate Если параметр установлен в {@code true}, то сервер автоматически завершает
   *     работу при первой же передышке в работе, когда нет ни одного открытого или ожидающего
   *     соединения, но после того, как уже было обработано хотя бы одно соединение.
   *     Если параметр установлен в {@code false}, то сервер ожидает поступления новых соединений
   *     до тех пор, пока не будет вызван метод {@link #mayTerminate()}.
   * @see #mayTerminate()
   */
  public void setAutoTerminate(boolean autoTerminate) {
    this.autoTerminate = autoTerminate;
  }

  public boolean getAutoTerminate() {
    return autoTerminate;
  }

  public void setDebugLog(OutputStream logStream) {
    this.messageLog = new PrintWriter(logStream);
    this.logFormatter = new XmlFormatter();
  }

  // ------------------------------------------------------
  // Главный цикл обработки

  // synchronized
  protected final Map<Thread, ConnectionHandler> threads = new HashMap<Thread, ConnectionHandler>();

  // synchronized
  private Queue<AbstractTraceMessage> messages = new LinkedList<AbstractTraceMessage>();

  /**
   * Синхронизированный доступ. Взять из очереди сообщение.
   *
   * @return Очередное сообщение или {@code null}.
   */
  protected AbstractTraceMessage pollMessage() {
    synchronized (messages) {
      return messages.poll();
    }
  }

  /** Синхронизированный доступ. Добавить в очередь сообщение. */
  protected void enqueueMessage(AbstractTraceMessage msg) {
    synchronized (messages) {
      messages.offer(msg);
    }
  }

  @Override
  public void process() throws IOException, IllegalStateException, IllegalArgumentException {
    if (serverSock == null) {
      throw new IllegalStateException("Socket already closed");
    }
    boolean verdict = false;
    String traceDesc = "tcp:" + port;
    fireTraceStart(traceDesc);

    try {
      for (; ; ) {
        // Check for new connections
        if (serverSock != null) { // May be null after closeSocket()
          try {
            Socket sock = serverSock.accept();
            ConnectionHandler ch = new ConnectionHandler(sock);
            Thread thrd = new Thread(ch);
            synchronized (threads) {
              threads.put(thrd, ch);
            }
            System.err.println(
                "XmlSocketServer: new connection accepted on port "
                    + port /*+ ", count=" + threads.size()*/);
            thrd.start();
          } catch (SocketTimeoutException ste) {
            // TODO!
          }
        }

        // Process messages
        AbstractTraceMessage msg;
        while ((msg = pollMessage()) != null) {
          if (messageLog != null) {
            String str = null;
            switch (msg.getKind()) {
              case TRACE_MESSAGE:
                str = logFormatter.formatMessage((TraceMessage) msg);
                break;
              case TAG_MESSAGE:
                str = logFormatter.formatTagMessage((TagMessage) msg);
                break;
              case DEBUG_MESSAGE:
                str = logFormatter.formatDebugMessage((DebugMessage) msg);
                break;
              default:
            }
            if (str != null) {
              messageLog.print(str);
            }
          }

          switch (msg.getKind()) {
            case TAG_MESSAGE:
              fireTagMessage((TagMessage) msg);
              break;
            case TRACE_MESSAGE:
              fireMessage((TraceMessage) msg);
              break;
            default:
          }
        }

        if (shutdownFlag && threads.isEmpty()) {
          verdict = true;
          return; // ==> finally
        }
      } // End main loop

    } finally {
      fireTraceEnd(traceDesc, verdict);
      closeSocket();
    }
  }

  /** Закрыть серверный сокет и установить {@link #mayTerminate() флаг завершения}. */
  public synchronized void closeSocket() {
    mayTerminate();
    if (serverSock == null) {
      return;
    }
    try {
      serverSock.close();
    } catch (IOException ioe) {
      // Silently ignore
    }
    serverSock = null;
    if (messageLog != null) {
      messageLog.close();
      messageLog = null;
    }
  }

  /**
   * Установить флаг завершения. После вызова этого метода сервер продолжает принимать новые
   * соединения (если не был вызван метод {@link #closeSocket()}) и обрабатывать старые, но сразу же
   * после закрытия всех соединений прекращает свою работу, не ожидая новых.
   *
   * @see #setAutoTerminate(boolean)
   */
  public void mayTerminate() {
    shutdownFlag = true;
  }

  @Override
  protected void finalize() throws Throwable {
    try {
      closeSocket();
    } finally {
      super.finalize();
    }
  }

  // ------------------------------------------------------

  /**
   * Разбирает XML-трассу, приходящую по сетевому соединению, и по мере разбора целых сообщений
   * вставляет их в общую очередь сообщений.
   */
  protected class ConnectionHandler extends XmlSocketParser implements Runnable {

    protected ConnectionHandler(Socket sock) {
      super(sock);
      setChannelPrefix(Tracer.allocateChannelId());
      // fireTraceStart() и fireTraceEnd() пустые, но
      // setTraceDesc() полезно для внятного описания источника в исключениях
      setTraceDesc("tcp:" + XmlSocketServer.this.port + ':' + getChannelPrefix());
    }

    @Override
    protected void fireMessage(TraceMessage msg) {
      enqueueMessage(msg);
    }

    @Override
    protected void fireTagMessage(TagMessage msg) {
      enqueueMessage(msg);
    }

    // Начало и конец трассы становятся началом и концом канала
    @Override
    protected void fireTraceStart(String traceDesc) {}

    @Override
    protected void fireTraceEnd(String traceDesc, boolean verdict) {}

    @Override
    public void run() {
      try {
        process(); // Цикл разбора полной XML-трассы
      } catch (Exception ex) {
        ex.printStackTrace();
      }
      synchronized (threads) {
        threads.remove(Thread.currentThread());
        // join() никто не вызывает
      }
      System.err.println(
          "XmlSocketServer: connection closed on port " + XmlSocketServer.this.port
          /*+ ", count=" + threads.size()*/ );

      super.closeSocket();

      if (autoTerminate) {
        mayTerminate();
      }
    }
  }
}
