/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.provider;

import com.ctc.wstx.api.ReaderConfig;
import com.ctc.wstx.api.WstxInputProperties;
import com.ctc.wstx.cfg.InputConfigFlags;
import com.ctc.wstx.stax.MinimalInputFactory;
import com.unitesk.aspectrace.MessageStream;
import com.unitesk.aspectrace.TraceNode;
import com.unitesk.aspectrace.TraceProvider;
import com.unitesk.aspectrace.sampleApp.SummaryCollector;
import com.unitesk.aspectrace.tracer.XmlConstants;
import com.unitesk.aspectrace.tracer.XmlFormatter;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

/**
 * Провайдер трассы: трасса получается при разборе XML-потока, отформатированного с помощью {@link
 * XmlFormatter}. Работает в модели PULL, поэтому вся работа происходит внутри вызова метода {@link
 * #process()}.
 *
 * <p>Провайдер монолитный: отсутствует фаза {@link MessageStream потока сообщений}, в который можно
 * встраиваться, а сообщения сразу раздаются адаптерам.
 *
 * @author Sergey Groshev
 */
public abstract class XmlStreamParser extends TraceProvider implements XmlConstants {
  private ArrayList<TraceNode> nodeStack = new ArrayList<TraceNode>();

  private ArrayList<TraceNode> fsmSummaryStack = new ArrayList<TraceNode>();

  private Map<String, String> curAttrs = null;

  private boolean insideTag = false;
  private boolean insideTrace = false;

  private String defChannel;
  private String curChannel;

  private XMLStreamReader xmlReader = null;

  private boolean summaryFlag = false;

  private static MinimalInputFactory factory;

  private static synchronized MinimalInputFactory getFactory() {
    if (factory != null) {
      return factory;
    }

    factory = MinimalInputFactory.newMinimalInstance();
    factory.setProperty(
        WstxInputProperties.P_INPUT_PARSING_MODE, WstxInputProperties.PARSING_MODE_DOCUMENT);
    ReaderConfig config = factory.getConfig();
    config.setConfigFlags(
        InputConfigFlags.CFG_COALESCE_TEXT
            | InputConfigFlags.CFG_NORMALIZE_LFS
            | InputConfigFlags.CFG_PRESERVE_LOCATION
            // | InputConfigFlags.CFG_REPORT_CDATA // Does not matter
            | InputConfigFlags.CFG_AUTO_CLOSE_INPUT);
    config.clearConfigFlag(InputConfigFlags.CFG_REPORT_PROLOG_WS);
    // Important: allow namespaces without declaration
    config.clearConfigFlag(InputConfigFlags.CFG_NAMESPACE_AWARE);
    return factory;
  }

  public XmlStreamParser() {}

  protected abstract String getSysId();

  protected abstract InputStream getInputStream() throws IOException, IllegalStateException;

  protected abstract String getTraceDesc();

  /**
   * Преобразует идентификатор канала, полученный из входной трассы, в идентификатор канала для
   * использования в системе обработки трасс. Реализация по умолчанию возвращает полученное
   * значение.
   */
  protected String channelFor(String traceChannelId) {
    return traceChannelId;
  }

  // ------------------------------------------------------

  @Override
  public void process() throws IOException, IllegalStateException, IllegalArgumentException {
    if (xmlReader != null) {
      throw new IllegalStateException("Parsing already started");
    }
    defChannel = curChannel = null;

    String aspect = null;
    Exception parseException = null;
    try {
      xmlReader = getFactory().createXMLStreamReader(getSysId(), getInputStream());

      while (xmlReader.hasNext()) {
        TraceNode tn;
        int colonIndex;
        String lclName;
        String ns;
        String shortName;
        int depth = nodeStack.size();

        curAttrs = null;
        boolean fsmSummary = false;

        KIND_SWITCH:
        switch (xmlReader.next()) {
            // case XMLStreamConstants.START_DOCUMENT:
            // moved to start of tag "trace"
          case XMLStreamConstants.START_ELEMENT:
            // xmlReader.getName().getPrefix() - for proper namespace processing
            lclName = xmlReader.getLocalName();

            colonIndex = lclName.indexOf(':');
            ns = colonIndex < 0 ? null : lclName.substring(0, colonIndex);
            shortName = colonIndex < 0 ? lclName : lclName.substring(colonIndex + 1);

            if (insideTag) {
              throw new IllegalArgumentException("Tag node has child node(s)");
            }

            if (depth != 0 && ns != null) {
              throw new IllegalArgumentException(
                  "XML namespace in child node: " + lclName + " is illegal here");
            }
            if (depth == 0 && ns == null) { // Process some special messages
              if (TAGNAME_TRACE.equals(shortName)) {
                if (insideTrace) {
                  throw new IllegalArgumentException("Nested <trace> tag in XML trace");
                }
                defChannel = getXmlAttrs().get(ATTR_TRACE_DEFCHANNEL);
                if (defChannel == null) {
                  defChannel = "";
                }

                fireTraceStart(getTraceDesc());
                insideTrace = true;
                break KIND_SWITCH; // Do not add node to nodeStack
              }

              if (!insideTrace) {
                throw new IllegalArgumentException("<trace> tag expected in XML trace");
              }
              if (TAGNAME_TAG_START.equals(shortName)) {
                processTag(true);
                insideTag = true;
                break KIND_SWITCH; // Do not add node to nodeStack
              } else if (TAGNAME_TAG_END.equals(shortName)) {
                processTag(false);
                insideTag = true;
                break KIND_SWITCH; // Do not add node to nodeStack
              } else if (TAGNAME_SYNC.equals(shortName)) {
                processSyncTag();
                insideTag = true;
                break KIND_SWITCH; // Do not add node to nodeStack
              }
              // Other aspectless messages will be processed by CommonAdapter.
            }
            if (depth == 0) {
              String ch = getXmlAttrs().get(ATTR_CHANNEL);
              curChannel = ch != null ? ch : defChannel;

              aspect = ns != null ? ns : "";
            }

            // TODO: skip building node tree for aspects without adapter
            tn = new TraceNode(shortName, getXmlAttrs());
            addChildToLast(tn);
            nodeStack.add(tn);

            if (lclName.equals("property")) {
              fsmSummaryStack.add(tn);
              summaryFlag = true;
            }

            break;

          case XMLStreamConstants.CHARACTERS:
          case XMLStreamConstants.CDATA:
            String text = xmlReader.getText();
            if (text.trim().length() == 0) {
              break;
            }
            if (insideTag) {
              throw new IllegalArgumentException("CHARACTERS or CDATA inside tag node");
            }
            if (depth == 0) {
              throw new IllegalArgumentException("CHARACTERS or CDATA on top level in XML trace");
            }
            tn = TraceNode.createTextNode(text);
            addChildToLast(tn);
            break;

          case XMLStreamConstants.END_ELEMENT:
            lclName = xmlReader.getLocalName();

            if (lclName.equals("fsm:summary")) {
              SummaryCollector.addToSummaryCollector(getTraceDesc(), fsmSummaryStack);
            }

            if (summaryFlag == false) {
              SummaryCollector.addToSummaryCollector(getTraceDesc(), null);
            }

            colonIndex = lclName.indexOf(':');
            ns = colonIndex < 0 ? null : lclName.substring(0, colonIndex);
            shortName = colonIndex < 0 ? lclName : lclName.substring(colonIndex + 1);

            if (depth == 0 && ns == null) {
              if (TAGNAME_TRACE.equals(shortName)) {
                insideTrace = false;
                fireTraceEnd(getTraceDesc(), true);
                break KIND_SWITCH;
              }

              if (TAGNAME_TAG_START.equals(shortName)
                  || TAGNAME_TAG_END.equals(shortName)
                  || TAGNAME_SYNC.equals(shortName)) {
                insideTag = false;
                break KIND_SWITCH; // Was not added to nodeStack
              }
            }

            tn = nodeStack.remove(depth - 1);
            if (depth == 1) {
              fireMessage(aspect, channelFor(curChannel), tn);
            }
            break;
          default:

            // case XMLStreamConstants.END_DOCUMENT:
            // moved to finally clause
        }
      } // end XML reader loop

    } catch (XMLStreamException xse) { // Including wrapped IOException
      parseException = xse;
    } catch (IllegalArgumentException iae) {
      parseException = iae;
    } finally {
      finalizeParsing(parseException);
    }
  }

  private void finalizeParsing(Exception parseException) throws IllegalArgumentException {
    String loc = null;
    if (parseException != null && xmlReader != null) {
      // TODO: привязать исключения от адаптеров к TraceNode, а TraceNode - к позиции в файле
      javax.xml.stream.Location location = xmlReader.getLocation();
      loc =
          "  in "
              + getTraceDesc()
              + ", line "
              + location.getLineNumber()
              + ", column "
              + location.getColumnNumber();
    }

    try {
      if (xmlReader != null) {
        xmlReader.close();
      }
    } catch (XMLStreamException xse) {
      if (parseException == null) {
        parseException = xse;
      }
    }
    xmlReader = null;
    nodeStack.clear();
    if (insideTrace) { // Abnormal exit
      insideTrace = false;
      fireTraceEnd(getTraceDesc(), false);
    }

    if (parseException != null) {
      String text = parseException.getMessage();
      if (loc != null) {
        text = text + "\n" + loc;
      }
      throw new IllegalArgumentException(text, parseException);
    }
  }

  private void addChildToLast(TraceNode tn) {
    int pntIndex = nodeStack.size() - 1;
    if (pntIndex >= 0) {
      nodeStack.get(pntIndex).addChild(tn);
    }
  }

  /** Обработчик сообщений tag-start и tag-end. */
  private void processTag(boolean isStart) {
    Map<String, String> attrs = getXmlAttrs();
    String aspect = attrs.get(ATTR_TAG_ASPECT);
    String tag = attrs.get(ATTR_TAG_NAME);
    String parent = attrs.get(ATTR_TAG_PARENT);
    curChannel = attrs.get(ATTR_CHANNEL);
    if (curChannel == null) {
      curChannel = defChannel;
    }

    if (isStart) {
      addProviderTag(aspect, channelFor(curChannel), tag, parent);
    } else {
      removeProviderTag(aspect, channelFor(curChannel), tag, parent);
    }
  }

  /** Обработчик сообщения sync. */
  private void processSyncTag() {
    Map<String, String> attrs = getXmlAttrs();
    String aspect = attrs.get(ATTR_SYNC_ASPECT);
    String channel = attrs.get(ATTR_SYNC_CHANNEL);
    String name = attrs.get(ATTR_SYNC_NAME);
    fireSyncMessage(aspect, channel, name);
  }

  /**
   * Получить набор атрибутов (имя +> значение) текущего XML-тэга. Имена атрибутов полностью
   * квалифицированные.
   */
  private Map<String, String> getXmlAttrs() {
    if (curAttrs != null) {
      return curAttrs;
    }
    int count = xmlReader.getAttributeCount();
    if (count == 0) {
      curAttrs = Collections.emptyMap();
    } else {
      curAttrs = new HashMap<String, String>(count, 1.0f);
      for (int i = 0; i < count; i++) {
        curAttrs.put(xmlReader.getAttributeLocalName(i), xmlReader.getAttributeValue(i));
      }
    }
    return curAttrs;
  }

  public ArrayList<TraceNode> getFsmSummaryStack() {
    return fsmSummaryStack;
  }
}
