/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.requirement;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;
import com.unitesk.aspectrace.sampleApp.SummaryCollector;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;

/**
 * Адаптер для аспекта "requirement".
 *
 * @author Anastasia Tugaenko
 */
public class RequirementAdapter extends AspectAdapter {

  @Override
  public String getAspect() {
    return "requirement";
  }

  @Override
  public Class<RequirementListener> getListenerClass() {
    return RequirementListener.class;
  }

  @Override
  protected void processMessage(TraceMessage msg) throws IllegalArgumentException {
    super.processMessage(msg);

    String msgName = msg.getName();

    if (msgName.equals("start")) {
      RequirementStartMessage reqMsg = new RequirementStartMessage(this);
      reqMsg.name = msg.getAttr("name");
      for (RequirementListener lst : getListeners()) {
        lst.processRequirementStart(reqMsg);
      }
    } else if (msgName.equals("db")) {
      String id = msg.getAttr("dbid");
      String dbName = msg.getAttr("name");
      String location = msg.getAttr("location");

      // parsing db
      HashMap<String, DBEntry> base = new HashMap<String, DBEntry>();
      XmlBdParser xbp = new XmlBdParser(location);
      try {
        xbp.process(id, base);
      } catch (IllegalStateException ise) {
        ise.printStackTrace();
      } catch (IllegalArgumentException iae) {
        iae.printStackTrace();
      } catch (IOException ioe) {
        ioe.printStackTrace();
      }
      DB db = new DB(id, dbName, location, base);
      SummaryCollector.addToAllDBList(id, db);

      for (RequirementListener lst : getListeners()) {
        lst.processRequirementDb(id, dbName, location);
      }
    } else if (msgName.equals("pass")) {
      String id = msg.getAttr("qid");
      String name = msg.getAttr("name");
      String db = msg.getAttr("db");
      for (RequirementListener lst : getListeners()) {
        lst.processRequirementPassed(id, db);
      }
      String testName = SummaryCollector.getCurrentTestName();
      ReqInfo reqInfo;
      if (name != null) {
        reqInfo = new ReqInfo(name, db, "-", testName, "passed");
      } else {
        reqInfo = new ReqInfo(id, db, "-", testName, "passed");
      }
      SummaryCollector.addToDbList(reqInfo);

    } else if (msgName.equals("fail")) {

      String id = msg.getAttr("qid");
      String name = msg.getAttr("name");
      String db = msg.getAttr("db");

      String message = null;
      for (TraceNode n : msg.getChildren()) {
        message = n.getText();
        for (RequirementListener lst : getListeners()) {
          lst.processRequirementFailed(id, db, message);
        }
      }
      if (message != null) {
        String testName = SummaryCollector.getCurrentTestName();
        ReqInfo reqInfo;
        if (id != null) {
          reqInfo = new ReqInfo(id, db, "-", testName, "failed");
        } else {
          reqInfo = new ReqInfo(name, db, "-", testName, "failed");
        }
        SummaryCollector.addToDbList(reqInfo);

      } else {
        String testName = SummaryCollector.getCurrentTestName();
        ReqInfo reqInfo = new ReqInfo(id, db, "-", testName, "failed");
        SummaryCollector.addToDbList(reqInfo);
      }

    } else {
      throw new IllegalArgumentException("Unsupported test node: " + msgName);
    }
  }

  @Override
  protected List<RequirementListener> getListeners() {
    return (List<RequirementListener>) super.getListeners();
  }
}
