/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.requirement;

import com.ctc.wstx.api.ReaderConfig;
import com.ctc.wstx.api.WstxInputProperties;
import com.ctc.wstx.cfg.InputConfigFlags;
import com.ctc.wstx.stax.MinimalInputFactory;
import com.unitesk.aspectrace.TraceNode;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

public class XmlBdParser implements XmlReqDBConstants {

  private ArrayList<TraceNode> nodeStack = new ArrayList<TraceNode>();

  private boolean insideTrace = false;

  private XMLStreamReader xmlReader = null;

  private static MinimalInputFactory factory;

  private URL DBFile;

  public XmlBdParser() {}

  public XmlBdParser(String db) {
    setDtBs(db);
  }

  public XmlBdParser(URL db) {
    setDtBs(db);
  }

  public void setDtBs(String db) {
    try {
      File cwd = new File(".");
      URL base = cwd.toURI().toURL();
      setDtBs(new URL(base, db));
    } catch (MalformedURLException mue) {
      throw new RuntimeException(mue);
    }
  }

  public void setDtBs(URL db) {
    if (DBFile != null) {
      throw new IllegalStateException("DB already set");
    }
    DBFile = db;
  }

  protected InputStream getInputStream() throws IOException {
    return new BufferedInputStream(DBFile.openStream());
  }

  protected String getSysId() {
    return DBFile.toString();
  }

  protected String getTraceDesc() {
    return getSysId();
  }

  private static synchronized MinimalInputFactory getFactory() {
    if (factory != null) {
      return factory;
    }

    factory = MinimalInputFactory.newMinimalInstance();
    factory.setProperty(
        WstxInputProperties.P_INPUT_PARSING_MODE, WstxInputProperties.PARSING_MODE_DOCUMENT);
    ReaderConfig config = factory.getConfig();
    config.setConfigFlags(
        InputConfigFlags.CFG_COALESCE_TEXT
            | InputConfigFlags.CFG_NORMALIZE_LFS
            | InputConfigFlags.CFG_PRESERVE_LOCATION
            // | InputConfigFlags.CFG_REPORT_CDATA // Does not matter
            | InputConfigFlags.CFG_AUTO_CLOSE_INPUT);
    config.clearConfigFlag(InputConfigFlags.CFG_REPORT_PROLOG_WS);
    // Important: allow namespaces without declaration
    config.clearConfigFlag(InputConfigFlags.CFG_NAMESPACE_AWARE);
    return factory;
  }

  public void process(String dbName, HashMap<String, DBEntry> base)
      throws IOException, IllegalStateException, IllegalArgumentException {
    if (xmlReader != null) {
      throw new IllegalStateException("Parsing already started");
    }

    Exception parseException = null;
    try {
      xmlReader = getFactory().createXMLStreamReader(getSysId(), getInputStream());
      String currentTag = "";

      String currentQid = "";
      String currentName = "";
      String currentDesc = "";
      String currentLdesc = "";

      boolean isFirst = true;
      while (xmlReader.hasNext()) {

        switch (xmlReader.next()) {
          case XMLStreamConstants.START_ELEMENT:
            currentTag = xmlReader.getLocalName();

            if (ATTR_DB_REQUIREMENT.equals(currentTag)) {

              if (!isFirst) {
                DBEntry dbEntry =
                    new DBEntry(dbName, currentQid, currentName, currentDesc, currentLdesc);
                base.put(currentQid, dbEntry);
                currentName = "";
                currentDesc = "";
                currentLdesc = "";
              }

              currentQid = xmlReader.getAttributeValue(0);
              if (isFirst) {
                isFirst = false;
              }
            }
            break;

          case XMLStreamConstants.CHARACTERS:
            if (ATTR_DB_REQUIREMENT_NAME.equals(currentTag)) {
              String temp = xmlReader.getText();
              currentName += temp.replaceAll("\\n", "<br>");
            }
            break;
          case XMLStreamConstants.CDATA:
            if (ATTR_DB_DESC.equals(currentTag)) {
              currentDesc += xmlReader.getText().replaceAll("\\n", "<br>");
            }
            if (ATTR_DB_LDESC.equals(currentTag)) {
              currentLdesc += xmlReader.getText().replaceAll("\\n", "<br>");
            }
            break;

          case XMLStreamConstants.END_ELEMENT:
            currentTag = xmlReader.getLocalName();
            if (TAGNAME_REQUALITY.equals(currentTag)) {
              DBEntry dbEntry =
                  new DBEntry(dbName, currentQid, currentName, currentDesc, currentLdesc);
              base.put(currentQid, dbEntry);
            }
            break;
          default:
        }
      } // end XML reader loop

    } catch (XMLStreamException xse) { // Including wrapped IOException
      parseException = xse;
    } catch (IllegalArgumentException iae) {
      parseException = iae;
    } finally {
      finalizeParsing(parseException);
    }
  }

  private void finalizeParsing(Exception parseException) throws IllegalArgumentException {
    String loc = null;
    if (parseException != null && xmlReader != null) {
      // TODO: привязать исключения от адаптеров к TraceNode, а TraceNode - к позиции в файле
      javax.xml.stream.Location location = xmlReader.getLocation();
      loc =
          "  in "
              + getTraceDesc()
              + ", line "
              + location.getLineNumber()
              + ", column "
              + location.getColumnNumber();
    }

    try {
      if (xmlReader != null) {
        xmlReader.close();
      }
    } catch (XMLStreamException xse) {
      if (parseException == null) {
        parseException = xse;
      }
    }
    xmlReader = null;
    nodeStack.clear();
    if (insideTrace) { // Abnormal exit
      insideTrace = false;
    }

    if (parseException != null) {
      String text = parseException.getMessage();
      if (loc != null) {
        text = text + "\n" + loc;
      }
      throw new IllegalArgumentException(text, parseException);
    }
  }
}
