/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.sampleApp;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.TagMessage;
import com.unitesk.aspectrace.coverage.CoverageListener;
import com.unitesk.aspectrace.coverage.CoverageMessage;
import com.unitesk.aspectrace.coverage.CoverageStructureMessage;
import com.unitesk.aspectrace.html.HeaderContext;
import com.unitesk.aspectrace.html.HtmlReportPage;
import com.unitesk.aspectrace.html.ReportGenerator;
import com.unitesk.aspectrace.html.ReportPage;
import com.unitesk.aspectrace.html.StaticResource;
import com.unitesk.aspectrace.ref.ReferenceResolver;
import com.unitesk.aspectrace.templates.DocumentContext;
import com.unitesk.coverage.CollectableCoverageStructure;
import com.unitesk.coverage.CoverageElement;
import com.unitesk.coverage.CoverageStructure;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.runtime.RuntimeConstants;
import org.apache.velocity.runtime.resource.loader.ClasspathResourceLoader;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

/**
 * Collects coverage information. For each leaf element having been reached, stores the hit count
 * and the context of the first hit.
 *
 * <p>For each coverage structure, generates the page and resolves references to it (via {@link
 * ReferenceResolver}) with parameters:<br>
 * <code>aspect="coverage"</code> и <code>ref=targetAspect + ":" + coverageName</code>
 *
 * @author Sergey Groshev
 */
public class CoverageCollectorCTESK
    implements ConfigurableGenerator, CoverageListener, ReferenceResolver<HtmlReportPage> {
  public static final String PROP_GENERATE_HITS = "CoverageCollector.generate.hits";
  public static final String PROP_GENERATE_FIRST_COVERED =
      "CoverageCollector.generate.first.covered";
  public static final String PROP_GENERATE_PINERY = "CoverageCollector.generate.pinery";

  /** {@code [aspect +> [struct name +> struct]]}. */
  protected Map<String, LinkedHashMap<String, CoverageStructure>> structs =
      new HashMap<String, LinkedHashMap<String, CoverageStructure>>();
  /** {@code [covered element +> context of first hit]}. */
  protected Map<CoverageElement, Map<String, Set<String>>> firstCoveredEnv =
      new HashMap<CoverageElement, Map<String, Set<String>>>();

  protected Map<CoverageElement, String> firstCoveredInChannel =
      new HashMap<CoverageElement, String>();

  protected boolean configHits = true;
  protected boolean configFirstCovered = false;
  protected boolean configPinery = false;

  public CoverageCollectorCTESK() {}

  public CoverageCollectorCTESK(Properties config) {
    configure(config);
  }

  @Override
  public void configure(Properties config) {
    configHits = Boolean.parseBoolean(config.getProperty(PROP_GENERATE_HITS, "true"));
    configFirstCovered =
        Boolean.parseBoolean(config.getProperty(PROP_GENERATE_FIRST_COVERED, "false"));
    configPinery = Boolean.parseBoolean(config.getProperty(PROP_GENERATE_PINERY, "false"));
  }

  // ------------------------------------------------------
  // CoverageListener

  @Override
  public void processCoverageStructure(CoverageStructureMessage csm) {
    CoverageStructure cs = csm.getCovStruct();
    String aspect = cs.getTargetAspect();
    LinkedHashMap<String, CoverageStructure> aspectStructs = structs.get(aspect);
    if (aspectStructs == null) {
      aspectStructs = new LinkedHashMap<String, CoverageStructure>();
      structs.put(aspect, aspectStructs);
    }
    if (aspectStructs.containsKey(cs.getName())) {

      // TODO: check equivalence of duplicates.
      // System.err.println("Coverage structure already present");
    } else {
      aspectStructs.put(cs.getName(), cs);
    }
  }

  /** Gets (probably, creates) the coverage element. */
  protected CoverageElement getCE(CoverageMessage cm) {
    String aspect = cm.getTargetAspect();
    LinkedHashMap<String, CoverageStructure> aspectStructs = structs.get(aspect);
    if (aspectStructs == null) {
      aspectStructs = new LinkedHashMap<String, CoverageStructure>();
      structs.put(aspect, aspectStructs);
    }
    String structName = cm.getStructName();
    CoverageStructure cs = aspectStructs.get(structName);
    CollectableCoverageStructure ccs = null;
    if (cs == null) {
      cs = ccs = new CollectableCoverageStructure(aspect, structName);
      aspectStructs.put(structName, ccs);
    } else if (cs instanceof CollectableCoverageStructure) {
      ccs = (CollectableCoverageStructure) cs;
    }

    CoverageElement ce;
    if (ccs != null) {
      ce = ccs.addElement(true, cm.getElementChain());
    } else { // cs != null
      ce = cs.findElement(cm.getElementChain());
      if (ce == null) {
        throw new IllegalArgumentException(
            "Unknown coverage element: "
                + Arrays.toString(cm.getElementChain())
                + " in coverage structure \""
                + cs.getName()
                + "\" for aspect "
                + aspect);
      }
    }
    return ce;
  }

  @Override
  public void processCoverageMessage(CoverageMessage cm) {
    Map<String, Set<String>> context = configFirstCovered ? cm.getAllTags() : null;
    CoverageElement ce = getCE(cm);
    CoverageElement ce2 = ce;
    while (ce2 != null) {
      if (!firstCoveredEnv.containsKey(ce2) // First hit
          || context != null && firstCoveredEnv.get(ce2) == null) {
        firstCoveredEnv.put(ce2, context);
        firstCoveredInChannel.put(ce2, cm.getChannel());
      }
      ce2 = ce2.getParent();
      context = null;
    }

    // Will access firstCoveredEnv - already updated
    getPage(ce.getStructure()).getLeaf(ce).hits++;
  }

  @Override
  public void processTagChange(TagMessage msg) {}

  @Override
  public void attachedToAdapter(String aspect, AspectAdapter adapter) {}

  // ------------------------------------------------------
  // Page generation

  @Override
  public ReportPage[] getPages() {
    // Create pages for the coverage structures not having been hit.
    for (Map<String, CoverageStructure> map : structs.values()) {
      for (CoverageStructure cs : map.values()) {
        getPage(cs);
      }
    }

    ArrayList<ReportPage> pages = new ArrayList<ReportPage>();
    pages.addAll(this.pages.values());
    pages.addAll(this.leaves.values());
    if (!pages.isEmpty()) {
      pages.add(new StaticResource(MainApp.CSS_SRC, MainApp.CSS));
    }
    return pages.toArray(new ReportPage[pages.size()]);
  }

  protected static CoverageElement normalize(CoverageElement ce) {
    if (ce.getName() == null) {
      return ce.getParent();
    } else {
      return ce;
    }
  }

  private Map<CoverageStructure, CoveragePage> pages =
      new LinkedHashMap<CoverageStructure, CoveragePage>();
  private Map<CoverageElement, Leaf> leaves = new HashMap<CoverageElement, Leaf>();

  /** Get or create page. */
  protected CoveragePage getPage(CoverageStructure cs) {
    CoveragePage cp = pages.get(cs);
    if (cp != null) {
      return cp;
    }
    cp = new CoveragePage(cs);
    pages.put(cs, cp);
    return cp;
  }

  public class CoveragePage extends HtmlReportPage {

    CoverageStructure cs;

    protected CoveragePage() {}

    protected CoveragePage(CoverageStructure cs) {
      this.cs = cs;

      String path = cs.getTargetAspect() + "-" + cs.getName() + ".html";
      path = path.replaceAll("[<>:/]", "_");
      setPath("coverage/" + path);

      setTitle("Coverage report: " + cs.getTargetAspect() + " :: " + cs.getName());
      setStyle(MainApp.CSS);

      buildSubPages(cs.getElements());
    }

    /** Get or create leaf. */
    protected Leaf getLeaf(CoverageElement ce) {
      ce = normalize(ce);
      Leaf lf = leaves.get(ce);
      if (lf != null) {
        return lf;
      }

      lf = new Leaf(this, ce);
      leaves.put(ce, lf);
      return lf;
    }

    /* Is called from the constructor. It is assumed that coverage is fully collected. */
    void buildSubPages(CoverageElement[] elems) {
      for (int i = 0; i < elems.length; i++) {
        CoverageElement elem = elems[i];
        if (firstCoveredEnv.containsKey(normalize(elem))) {
          if (!elem.hasChildren()) {
            getLeaf(elem); // Will create it if needed
          } else {
            buildSubPages(elem.getExtendedChildren());
          }
        }
      }
    }

    // Maybe, will be used for auto-referencing later
    @Override
    public String getAspect() {
      return "coverage";
    }

    protected int coveredLeafCount;
    protected int hitCount;
    private File reportDir;

    @Override
    public void generatePageBody(PrintWriter pw) {

      pw.println("<h1>" + getTitle() + "</h1>");
      pw.println("<p>Generated: " + Calendar.getInstance().getTime() + "</p>");
      DocumentContext doc = new DocumentContext();
      try {
        Properties props = new Properties();
        props.put(RuntimeConstants.RESOURCE_LOADER, "classpath");
        props.put("classpath.resource.loader.class", ClasspathResourceLoader.class.getName());
        Velocity.init(props);
        VelocityContext vc = new VelocityContext();
        String templateName = "overview-failure.vm";
        Template tmplt = Velocity.getTemplate(doc.getTemplatePackage() + templateName);
        HeaderContext docCont = new HeaderContext(getGenerator(), this);
        SummaryCollector sc = new SummaryCollector();

        vc.put("doc", doc);
        vc.put("current", this);
        vc.put("docContext", docCont);
        vc.put("summary", sc);

        tmplt.merge(vc, pw);

        ReportGenerator rg = new ReportGenerator();
        reportDir = rg.getGenerationDir();

      } catch (Exception ex) {
        ex.printStackTrace();
      }

      coveredLeafCount = hitCount = 0;

      CoverageElement[] elements = cs.getElements();
      int maxDepth = getMaxDepth(elements);
      int width = maxDepth + (configHits ? 1 : 0) + (configFirstCovered ? 1 : 0);

      pw.println("<b>Aspect:</b> " + cs.getTargetAspect() + "<br/>");
      pw.println("<b>Coverage name:</b> " + encodeSimple(cs.getName()));

      pw.println("<table width=\"50%\" border=\"0\" class=\"border\" cellspacing=\"0\"> ");

      if (configHits || configFirstCovered) {
        pw.println("<tr><th colspan=\"" + maxDepth + "\">Coverage elements");
        if (configHits) {
          pw.println("<th style='text-align: center;'>Hits");
        }
        if (configFirstCovered) {
          pw.println("<th>First covered");
        }
      }
      pw.println("<tr>");
      generateRows(pw, elements, maxDepth);

      pw.println("<tr><th colspan=\"" + width + "\">");
      pw.print("Total leaf coverage: <b>" + coveredLeafCount + " / ");
      String coverageString = null;
      coverageString = coveredLeafCount + " / ";

      if (cs instanceof CollectableCoverageStructure) {
        pw.print("<font color=\"#FF0000\">?</font></b>");
        coverageString += "<font color=\"#FF0000\">?</font></b>";
      } else {
        pw.print(getHeight(cs.getElements()));
        coverageString += getHeight(cs.getElements());
      }
      if (configHits) {
        pw.print("<br>Hits: " + hitCount);
      }
      pw.println("</b>");

      SummaryCollector.setTotalCoverage(coverageString);

      pw.println("</table><br/>");
    }

    protected void generateRows(PrintWriter pw, CoverageElement[] elements, int colspan) {

      for (int i = 0; i < elements.length; i++) {
        CoverageElement ce = elements[i];
        CoverageElement[] children = ce.getExtendedChildren();
        int height = getHeight(children);
        boolean isCovered = firstCoveredEnv.containsKey(normalize(ce));
        String tdClass =
            !isCovered ? "uncovered" : hasUncoveredLeaves(ce) ? "semicovered" : "covered";

        String name = ce.getName();
        if (name == null || name.length() == 0) { // null for dummy child
          name = "&minus;";
        }

        if (i > 0) {
          pw.println("<tr>");
        }

        pw.print("<td class=\"" + tdClass + "\"");
        if (height > 1) {
          pw.print(" rowspan=" + height);
        }
        if (children.length == 0 && colspan > 1) {
          pw.print(" colspan=" + colspan);
        }
        pw.print(">");

        pw.print(name);
        pw.println("</td>");

        if (children.length == 0) {
          if (isCovered) {
            Leaf leaf = getLeaf(ce);
            coveredLeafCount++;
            if (configHits) {
              pw.print("<td class=\"" + tdClass + "\"align=center>" + leaf.hits);
              hitCount += leaf.hits;
            }
            if (configFirstCovered) {
              pw.print("<td class=\"" + tdClass + "\">&nbsp;");
              leaf.generateRefs(pw, true);
            }
          } else {
            if (configHits) {
              pw.print("<td class=\"uncovered\" align=\"center\">0");
            }
            if (configFirstCovered) {
              pw.print("<td class=\"uncovered\">&nbsp;&minus;");
            }
          }
          pw.println("</tr>");
        } else {
          generateRows(pw, children, colspan - 1);
        }
      }
    }

    protected boolean hasUncoveredLeaves(CoverageElement ce) {
      if (!firstCoveredEnv.containsKey(normalize(ce))) {
        return true;
      }

      CoverageElement[] children = ce.getExtendedChildren();
      for (int i = 0; i < children.length; i++) {
        if (hasUncoveredLeaves(children[i])) {
          return true;
        }
      }
      return false;
    }

    protected int getHeight(CoverageElement[] elems) {
      if (elems.length == 0) {
        return 1;
      }

      int sum = 0;
      for (int i = 0; i < elems.length; i++) {
        sum += getHeight(elems[i].getExtendedChildren());
      }
      return sum;
    }

    protected int getMaxDepth(CoverageElement[] elems) {
      int maxDepth = 1;
      for (int i = 0; i < elems.length; i++) {
        CoverageElement[] children = elems[i].getChildren();
        if (children.length > 0) {
          int depth = getMaxDepth(children) + 1;
          if (maxDepth < depth) {
            maxDepth = depth;
          }
        }
      }
      return maxDepth;
    }

    public int getLevel() {
      return 1;
    }
  } // end class CoveragePage

  protected class Leaf extends HtmlReportPage {

    int hits = 0;

    Leaf(CoveragePage cp, CoverageElement ce) {
      if (configFirstCovered) { // For auto-referencing
        Map<String, Set<String>> marks = firstCoveredEnv.get(ce);
        setTagMarks(marks);
      }

      setParent(cp, null);
      setAssociatedChannel(firstCoveredInChannel.get(ce));
    }

    // For auto-referencing
    @Override
    public String getAspect() {
      return "coverage";
    }

    @Override
    public void generatePageBody(PrintWriter pw) {}

    @Override
    public void generateRefs(PrintWriter pw, boolean compact) {
      if (configPinery) {
        super.generateRefs(pw, compact);
        Set<String> pineryTags = getTagMarks().get("pinery");
        if (pineryTags == null) {
          return;
        }
        pw.print("<b>");
        for (String s : pineryTags) {
          pw.print(s);
          pw.print(" ");
        }
        pw.println("</b>");
        return;
      }
      if (pageRefs.isEmpty()) {
        pw.println("<i>no context</i>");
      } else {
        super.generateRefs(pw, compact);
      }
    }
  } // end class Leaf

  // ------------------------------------------------------
  // ReferenceResolver

  // For tuning method resolve(String)
  protected HtmlReportPage resolve(CoverageStructure cs) {
    return getPage(cs); // will create page if needed
  }

  @Override
  public HtmlReportPage resolve(String channel, String ref) {
    int colonIndex = ref.indexOf(':');
    if (colonIndex < 0 || colonIndex == ref.length() - 1) {
      return null;
    }
    String targetAspect = ref.substring(0, colonIndex);
    Map<String, CoverageStructure> aspectStructs = structs.get(targetAspect);
    if (aspectStructs == null) {
      return null;
    }
    String csName = ref.substring(colonIndex + 1);
    CoverageStructure cs = aspectStructs.get(csName);
    if (cs == null) {
      return null;
    }
    return resolve(cs);
  }
}
