/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.sampleApp;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.TagMessage;
import com.unitesk.aspectrace.error.ErrorListener;
import com.unitesk.aspectrace.error.FailureMessage;
import com.unitesk.aspectrace.html.HeaderContext;
import com.unitesk.aspectrace.html.HtmlReportPage;
import com.unitesk.aspectrace.html.ReportGenerator;
import com.unitesk.aspectrace.html.ReportPage;
import com.unitesk.aspectrace.html.StaticResource;
import com.unitesk.aspectrace.ref.ContextProvider;
import com.unitesk.aspectrace.templates.DocumentContext;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

public class ErrorCollector implements ErrorListener, ConfigurableGenerator {

  private List<FailurePage> failures = new ArrayList<FailurePage>();
  private List<ContextProvider<HtmlReportPage>> providers =
      new ArrayList<ContextProvider<HtmlReportPage>>();

  boolean configDebugTags = false;
  boolean configStack = true;
  public File reportDir;

  public ErrorCollector() {}

  public ErrorCollector(Properties config) {
    configure(config);
  }

  @Override
  public void configure(Properties config) {
    configDebugTags =
        Boolean.parseBoolean(config.getProperty("ErrorCollector.generate.debug.tags", "false"));
    configStack = Boolean.parseBoolean(config.getProperty("ErrorCollector.generate.stack", "true"));
  }

  @Override
  public void processFailure(FailureMessage msg) {
    FailurePage page = new FailurePage(failures.size(), msg);
    for (ContextProvider<HtmlReportPage> sp : providers) {
      HtmlReportPage[] sections = sp.getRelated(msg.getChannel(), page);
      if (sections != null) {
        page.addSections(sections);
      }
    }
    failures.add(page);
  }

  @Override
  public void processTagChange(TagMessage msg) {}

  @Override
  public void attachedToAdapter(String aspect, AspectAdapter adapter) {}

  public FailurePage getPage(int index) {
    return failures.get(index);
  }

  @Override
  public ReportPage[] getPages() {
    if (failures.isEmpty()) {
      return new HtmlReportPage[0];
    }

    ReportPage[] rp = new ReportPage[failures.size() + 1];
    failures.toArray(rp);
    rp[rp.length - 1] = new StaticResource(MainApp.CSS_SRC, MainApp.CSS);
    return rp;
  }

  public void addSectionProvider(ContextProvider<HtmlReportPage> sp) {
    providers.add(sp);
  }

  /**
   * Класс страницы отчёта об ошибке. Общедоступный, потому что с этим компонентом и этими
   * страницами другие компоненты ({@link FsmGraphProcessor}) работают напрямую.
   *
   * @see ErrorCollector#getPage(int)
   */
  public class FailurePage extends HtmlReportPage {

    FailureMessage msg;
    List<HtmlReportPage> sections = new ArrayList<HtmlReportPage>();

    FailurePage(int index, FailureMessage msg) {
      super("failure/Failure_" + (index + 1) + ".html");
      setTitle("Failure #" + (index + 1) + ": " + msg.getShortDesc());
      setStyle(MainApp.CSS);

      this.msg = msg;
      setTagMarks(msg.getAllTags()); // Это копия, она останется с нами
    }

    void addSections(HtmlReportPage[] sections) {
      for (HtmlReportPage section : sections) {
        section.setParent(this, null);
        this.sections.add(section);
      }
    }

    @Override
    public String getAspect() {
      return "error";
    }

    @Override
    public void generatePageBody(PrintWriter pw) {

      pw.println("<h1>" + getTitle() + "</h1>");

      DocumentContext doc = new DocumentContext();
      try {
        Velocity.init();
        VelocityContext vc = new VelocityContext();
        Template tmplt = Velocity.getTemplate(doc.getTemplatePackage() + "overview-failure.vm");
        HeaderContext docCont = new HeaderContext(getGenerator(), this);
        vc.put("doc", doc);
        vc.put("current", this);
        vc.put("docContext", docCont);

        SummaryCollector sc = new SummaryCollector();

        vc.put("summary", sc);

        tmplt.merge(vc, pw);

        ReportGenerator rg = new ReportGenerator();
        reportDir = rg.getGenerationDir();

      } catch (Exception ex) {
        ex.printStackTrace();
      }

      pw.println("<table class=\"section\" width=\"100%\">");
      pw.println("<tr><th colspan=\"2\">Failure information");
      pw.println(
          "<tr><td width=\"15%\"><b>Origin aspect</b><td width=\"85%\">" + msg.getOriginAspect());
      pw.println("<tr><td><b>Short desc<td>" + msg.getShortDesc());
      pw.println("<tr><td><b>Failure text<td>" + msg.getFullDesc());

      if (configStack && msg.getStackSize() > 0) {
        pw.println("<tr><th colspan=\"2\">Call stack");
        for (int i = 0; i < msg.getStackSize(); i++) {
          pw.println("<tr><td colspan=\"2\">" + msg.getStackElement(i));
        }
      }

      if (configDebugTags && !getTagMarks().isEmpty()) {
        pw.println("<tr><th colspan=\"2\">Tag context");
        for (Map.Entry<String, Set<String>> me : getTagMarks().entrySet()) {
          pw.println("<tr><td>" + me.getKey() + "<td>" + me.getValue());
        }
      }
      pw.println("</table>");
      if (sections.size() > 0) {
        pw.println("<br/><h2>Failure context:</h2>");
        for (HtmlReportPage p : sections) {
          p.generatePageBody(pw);
          pw.println("<br>");
        }
      }
    }

    public int getLevel() {
      return 1;
    }

    public String getReportType() {
      return "failure";
    }
  } // end class FailurePage
}
