/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.sampleApp;

import com.unitesk.aspectrace.AspectListener;
import com.unitesk.aspectrace.TraceProcessorManager;
import com.unitesk.aspectrace.html.HtmlReportPage;
import com.unitesk.aspectrace.html.ReportGenerator;
import com.unitesk.aspectrace.html.ReportPage;
import com.unitesk.aspectrace.ref.ReferenceResolver;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class GeneralGenerateReport {

  private static final Pattern ASPECT_REGEX = Pattern.compile("^aspect\\..*");
  private static final Pattern ASPECT_NAME_REGEX = Pattern.compile("^(aspect\\.[^.]+)\\.name$");

  static final String DEFAULT_PROPERTIES = "Generator.properties";

  public static final String CSS = "report.css";
  public static final String CSS_SRC = "/com/unitesk/aspectrace/sampleApp/report.css";

  public static void main(String[] args) {
    generateReport();
  }

  // ------------------------------------------------------

  public static void generateReport() {
    try {
      System.out.println(">>> Generating report...");

      Map<String, ReferenceResolver<? extends HtmlReportPage>> aspectNameToResolverInstance =
          new HashMap<String, ReferenceResolver<? extends HtmlReportPage>>();

      TraceProcessorManager tpm = new TraceProcessorManager();
      ReportGenerator rg = new ReportGenerator();
      rg.setDefaultCharset("windows-1251");

      ConfigParser config = new ConfigParser(DEFAULT_PROPERTIES);

      String traces = config.getValue("traces");
      System.out.println("Traces to process: " + traces);

      String delims = config.getValue("tracesDelim"); // разделитель для трасс в файле свойств
      String[] parsedTraces = traces.split(delims);

      // Добавляем входные трассы
      for (String trace : parsedTraces) {
        tpm.addTrace(new File(trace));
      }

      // Разбираем файл свойств
      ArrayList<String> pattAspectKeys = getKeysSetByPattern(config.info, ASPECT_REGEX);
      System.out.println("\nAspects and classes:");
      for (String strName : pattAspectKeys) {
        Matcher matcherName = ASPECT_NAME_REGEX.matcher(strName);
        if (matcherName.matches()) {
          String aspectPrefix = matcherName.group(1);

          String aspectName = config.getValue(strName, null);
          String processorName = config.getValue(aspectPrefix + ".processor", null);
          String listenerName = null;
          String resolverName = null;

          if (aspectName == null) {
            System.err.println("ERROR: Aspect name is not specified , prefix = " + aspectPrefix);
          }

          if (processorName == null) {
            listenerName = config.getValue(aspectPrefix + ".listener", null);
            resolverName = config.getValue(aspectPrefix + ".resolver", null);

            if (listenerName == null) {
              System.err.println(
                  "ERROR: Processor and listener for aspect " + aspectName + " are not specified");
            }

            if (resolverName == null) {
              System.err.println(
                  "ERROR: Processor and resolver for aspect " + aspectName + " are not specified");
            }
          }

          if (!((aspectName != null)
              || (processorName != null)
              || (resolverName != null)
              || (listenerName != null))) {
            System.out.println("Error in parsing property file, ");
          } else {

            if (processorName != null) {
              System.out.println(" processor = " + processorName);
              Class<?> cls = Class.forName(processorName);
              AspectListener listener = (AspectListener) cls.newInstance();
              tpm.addListener(aspectName, listener);
              rg.addRefResolver(aspectName, (ReferenceResolver<? extends HtmlReportPage>) listener);

              aspectNameToResolverInstance.put(
                  aspectName, (ReferenceResolver<? extends HtmlReportPage>) listener);
            } else {
              if (listenerName != null) {
                System.out.println(" listener = " + listenerName);
                Class<?> cls = Class.forName(listenerName);
                AspectListener listener = (AspectListener) cls.newInstance();
                tpm.addListener(aspectName, listener);

                if (resolverName == null) {
                  aspectNameToResolverInstance.put(
                      aspectName, (ReferenceResolver<? extends HtmlReportPage>) listener);
                }
              }
              if (resolverName != null) {
                System.out.println(" resolver = " + resolverName);
                Class<?> cls = Class.forName(resolverName);
                ReferenceResolver<? extends HtmlReportPage> resolver =
                    (ReferenceResolver<? extends HtmlReportPage>) cls.newInstance();
                rg.addRefResolver(aspectName, resolver);

                aspectNameToResolverInstance.put(aspectName, resolver);
              }
            }
          }
        }
      }

      // добавляем связи
      Pattern pattReference = Pattern.compile("reference\\..*?");
      ArrayList<String> pattRefKeys = getKeysSetByPattern(config.info, pattReference);

      System.out.println("References:");

      for (String str : pattRefKeys) {
        String ref = config.getValue(str);
        String[] fromToRef = ref.split("->");
        System.out.println(fromToRef[0] + " -> " + fromToRef[1]);
        rg.addAutoRef(fromToRef[0], fromToRef[1]);
      }

      // Запускаем обработку
      try {
        tpm.process();
      } catch (Exception ex) {
        ex.printStackTrace();
        System.err.println();
        System.err.println("Trying to generate report after failure...");
      }

      // TODO

      ////////// Собираем и генерируем HTML-отчёт //////////

      String reportDir = config.getValue("reportDirectory", null);
      if (reportDir == null) {
        reportDir = "./Report" + System.currentTimeMillis();
        System.out.println("INFO: Report directory is not defined, write in " + reportDir);
      } else {
        reportDir = reportDir + System.currentTimeMillis();
      }

      rg.setGenerationDir(new File(reportDir));

      String inPath = config.getValue("styleFilePath.in", null);
      String outPath = config.getValue("styleFilePath.out", null);
      if ((inPath == null) || (inPath == null)) {
        inPath = "/com/unitesk/aspectrace/sampleApp/report.css";
        outPath = "report.css";
        System.out.println(
            "INFO: Path to resource is not defined, got default" + inPath + " " + outPath);
      }

      rg.addResource(inPath, outPath);
      rg.addIndexPage();

      for (String aspectMapKey : aspectNameToResolverInstance.keySet()) {
        ReferenceResolver<? extends HtmlReportPage> instance =
            aspectNameToResolverInstance.get(aspectMapKey);

        Method getPagesReflect = instance.getClass().getMethod("getPages");

        Object obj = getPagesReflect.invoke(instance);
        ReportPage[] rp = (ReportPage[]) obj;
        System.out.println("rp = " + rp.length);
        rg.addPages(rp);
      }

      rg.generateReport();
      System.out.println("Report was generated");
    } catch (Throwable ex) {
      ex.printStackTrace();
    }
  }

  public static ArrayList<String> getKeysSetByPattern(Properties properties, Pattern patt) {

    String keyString = properties.keySet().toString();
    String[] keyArrays = keyString.split(", ");
    ArrayList<String> result = new ArrayList<String>();
    int keysNumber = keyArrays.length;
    int counter = 0;
    for (String raw : keyArrays) {
      counter++;
      if (counter == 1) {
        raw = raw.substring(1, raw.length());
      }
      if (counter == keysNumber) {
        raw = raw.substring(0, raw.length() - 1);
      }

      Matcher matcher = patt.matcher(raw);
      if (matcher.matches()) {
        result.add(raw);
      }
    }
    return result;
  }

  public static class ConfigParser {

    Properties info = new Properties();
    BufferedReader br;
    FileInputStream fin = null;

    public ConfigParser() {
      try {
        br = new BufferedReader(new InputStreamReader(System.in, "UTF-8"));
      } catch (UnsupportedEncodingException uee) {
        uee.printStackTrace();
      }
    }

    public ConfigParser(String name) {
      try {
        br = new BufferedReader(new InputStreamReader(System.in, "UTF-8"));
        readFromFile(name);
      } catch (UnsupportedEncodingException uee) {
        uee.printStackTrace();
      }
    }

    public void readFromFile(String name) {
      try {
        fin = new FileInputStream(name);
      } catch (FileNotFoundException fnfe) {
        System.out.println("file " + name + " not found");
      }

      try {
        if (fin != null) {
          info.load(fin);
          fin.close();
        }
      } catch (IOException ioe) {
        System.out.println("Error reading file.");
      }
    }

    public String getValue(String key) {
      return info.getProperty(key);
    }

    public String getValue(String key, String defaultValue) {
      return info.getProperty(key, defaultValue);
    }

    public Enumeration<Object> getKeys() {
      return info.keys();
    }
  }
}
