/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.sampleApp;

import com.unitesk.aspectrace.Parameter;
import com.unitesk.aspectrace.html.HtmlReportPage;
import com.unitesk.aspectrace.html.ReportPage;
import com.unitesk.aspectrace.html.StaticResource;
import com.unitesk.aspectrace.oracle.AbstractOracleListener;
import com.unitesk.aspectrace.oracle.MarkMessage;
import com.unitesk.aspectrace.oracle.MarkedPath;
import com.unitesk.aspectrace.oracle.OracleEndMessage;
import com.unitesk.aspectrace.oracle.OracleStartMessage;
import com.unitesk.aspectrace.oracle.OracleTracer;
import com.unitesk.aspectrace.ref.ContextProvider;
import com.unitesk.aspectrace.ref.ReferenceResolver;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

public class OracleCollector extends AbstractOracleListener
    implements ConfigurableGenerator,
        ReferenceResolver<HtmlReportPage>,
        ContextProvider<HtmlReportPage> {
  private static final String COVERAGE_ANCHOR = "coverage";

  Map<String, OraclePage> oracles = new HashMap<String, OraclePage>();
  OracleCallPage currentCall = null;

  private boolean configIncludeCoverage = true;
  private boolean configLinkCoverage = false;

  public OracleCollector() {}

  public OracleCollector(Properties config) {
    configure(config);
  }

  @Override
  public void configure(Properties config) {
    String coverageProp = config.getProperty("OracleCollector.coverages");
    if (coverageProp != null) {
      configIncludeCoverage = "include".equalsIgnoreCase(coverageProp);
      configLinkCoverage = "link".equalsIgnoreCase(coverageProp);
    }
  }

  @Override
  public void processOracleStart(OracleStartMessage msg) {
    String signature = msg.getSignature();
    OraclePage op = oracles.get(signature);
    if (op == null) {
      op = new OraclePage(signature);
      oracles.put(signature, op);
    }
    op.counter++;

    // Just replace. If some error needs appropriate section, it has already
    // saved link to it.
    currentCall = new OracleCallPage(msg.getSignature(), msg.getParameters());
  }

  @Override
  public void processOracleEnd(OracleEndMessage msg) {
    currentCall = null;
  }

  @Override
  public void processMark(MarkMessage msg) {
    // Одинаково обрабатываем и branch, и mark
    currentCall.marks.add(msg.getName());
  }

  @Override
  public ReportPage[] getPages() {
    if (oracles.isEmpty()) {
      return new ReportPage[0];
    }

    ReportPage[] rp = new ReportPage[oracles.size() + 1];
    oracles.values().toArray(rp);
    rp[rp.length - 1] = new StaticResource(MainApp.CSS_SRC, MainApp.CSS);
    return rp;
  }

  protected class OraclePage extends HtmlReportPage {
    int counter = 0;
    String signature;
    List<HtmlReportPage> coverageSections;

    public OraclePage(String signature) {
      super("oracle/" + signature.replace('<', '[').replace('>', ']') + ".html");
      setTitle("Oracle " + signature);
      setStyle(MainApp.CSS);

      this.signature = signature;
    }

    @Override
    public void buildRefs() {
      addCoverageSection(getGenerator().resolve("coverage", null, "oracle:" + signature));

      addCoverageSection(
          getGenerator().resolve("coverage", null, "oracle:" + MarkedPath.CS_PREFIX + signature));

      addCoverageSection(getGenerator().resolve("coverage", null, "oracle:disjunct:" + signature));
    }

    private void addCoverageSection(HtmlReportPage section) {
      if (section == null) {
        return;
      }

      if (configIncludeCoverage) { // Include coverage page
        if (coverageSections == null) {
          coverageSections = new ArrayList<HtmlReportPage>();
        }
        coverageSections.add(section);
        section.setParent(this, COVERAGE_ANCHOR);
      } else if (configLinkCoverage) { // Cross-reference
        addPageRef(section, null);
        section.addPageRef(this, null);
      }
    }

    @Override
    public void generatePageBody(PrintWriter pw) {
      pw.println("<h1>Oracle info: " + signature + "</h1>");
      pw.println("Call count: " + counter);
      if (coverageSections != null) {
        pw.print("<h2><a name=\"" + COVERAGE_ANCHOR + "\">");
        pw.println("Coverage reports:</a></h2>");
        for (HtmlReportPage p : coverageSections) {
          p.generatePageBody(pw);
        }
      }
    }

    @Override
    public String generateRef(HtmlReportPage from, String mark) {
      if (mark != null && mark.equals(signature)) {
        mark = null; // Strip mark after resover
      }
      return super.generateRef(from, mark);
    }
  } // end class OraclePage

  // --------------- ContextProvider

  @Override
  public HtmlReportPage[] getRelated(String channel, HtmlReportPage target) {
    if (currentCall == null) {
      return null;
    } else {
      return new HtmlReportPage[] {currentCall};
    }
  }

  // TODO: externalize
  static String[] showScript = {
    "<script type=\"text/javascript\"><!--",
    "function show(id,al) {",
    " el=document.getElementById(id);",
    " if(el.style.display != 'none') {",
    "  el.style.display = 'none';",
    "  al.innerHTML = '+';",
    " } else {",
    "  el.style.display = '';",
    "  al.innerHTML = '&#150;';",
    " }",
    " return false;",
    "}",
    "--></script>"
  };

  /**
   * Контекстная секция информации о вызове. Особенности реализации:
   *
   * <ul>
   *   <li>Если секцию запросили 0 раз - она теряется сразу после завершения вызова оракула и
   *       экономит память.
   *   <li>Если запросили 1 раз - становится честной секцией запросившей страницы, которая и хранит
   *       на неё ссылку.
   *   <li>Если запросили 2 и более раз - ссылки сохраняются, но секция может ссылаться на
   *       неправильную родительскую страницу. Для преодоления этой потенциальной проблемы надо
   *       настроить своевременное клонирование.
   * </ul>
   */
  protected static class OracleCallPage extends HtmlReportPage implements Cloneable {
    String signature;
    Parameter[] params;
    List<String> marks = new ArrayList<String>();

    public OracleCallPage(String signature, Parameter[] parameters) {
      super(); // null path, never create separate file
      this.signature = signature;
      this.params = parameters;
    }

    @Override
    public void generatePageBody(PrintWriter pw) {
      for (int i = 0; i < showScript.length; i++) {
        pw.println(showScript[i]);
      }
      pw.println("<table class=\"section\" width=\"100%\">");
      pw.println("<tr><th colspan=\"2\">Test situation</th></tr>");

      HtmlReportPage oraPage = getGenerator().resolve("oracle", null, signature);
      String head = oraPage == null ? signature : oraPage.generateRef(this, null);
      pw.println(
          "<tr><td width=\"15%\"><b>Oracle</b></td><td width=\"85%\">" + head + "</td></tr>");

      pw.println("<tr><td colspan=\"2\"><b>Call parameters:</b></td></tr>");
      for (Parameter p : params) {
        pw.print("<tr><td colspan=\"2\">");
        pw.print(
            encodeMultiLine(
                encodeSimple(p.getType())
                    + " <b>"
                    + encodeSimple(p.getName())
                    + "</b> = "
                    + encodeSimple(p.getValue())));
        pw.println("</td></tr>");
      }

      if (marks.size() > 0) {
        String markInfo = marks.toString();
        HtmlReportPage covPage =
            getGenerator()
                .resolve(
                    "coverage", null, OracleTracer.ASPECT + ':' + MarkedPath.CS_PREFIX + signature);
        if (covPage != null) {
          // coverageSections строятся во время вызова buildRefs(), а
          // генерация секции происходит уже во время генерации отчёта.
          markInfo = covPage.generateTextRef(this, null, markInfo);
        }
        pw.println("<tr><td><b>Marked path</b></td>");
        pw.println("<td>" + markInfo + "</td></tr>");
      }
      pw.println("</table>");
    }

    private static int paramNum = 0;

    private static String encodeMultiLine(String str) {
      int idx = str.indexOf('\n');
      if (idx >= 0) {
        paramNum++;
        String uid = "id" + paramNum;
        return "<code>[<a onclick=\"return show('"
            + uid
            + "', this)\" href=\"#\" class=\"expand\">+</a>]"
            + str.substring(0, idx)
            + "</code>"
            + "<pre id=\""
            + uid
            + "\" style=\"display:none\" class=\"expand\">"
            + str.substring(idx + 1)
            + "</pre>";
      } else {
        return "<code>" + str + "</code>";
      }
    }
  } // end class OracleCallPage

  // --------------- ReferenceResolver

  @Override
  public HtmlReportPage resolve(String channel, String ref) {
    return oracles.get(ref);
  }
}
