/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.sampleApp;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.TagMessage;
import com.unitesk.aspectrace.html.HtmlReportPage;
import com.unitesk.aspectrace.html.ReportPage;
import com.unitesk.aspectrace.html.StaticResource;
import com.unitesk.aspectrace.ref.ContextProvider;
import com.unitesk.aspectrace.ref.ReferenceResolver;
import com.unitesk.aspectrace.requirement.DBReportPage;
import com.unitesk.aspectrace.requirement.RequirementListener;
import com.unitesk.aspectrace.requirement.RequirementStartMessage;

import java.io.PrintWriter;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

public class RequirementProcessor
    implements RequirementListener,
        ReferenceResolver<HtmlReportPage>,
        ContextProvider<HtmlReportPage>,
        ConfigurableGenerator {

  public static final String PROP_TEST_TRACING = "RequirementProcessor.testTagTracing";

  protected boolean configTestTracing = true;

  int reqNum = 0;

  AspectAdapter adapter;
  String reqName;
  String message;

  public String getMessage() {
    return message;
  }

  public RequirementProcessor() {}

  public RequirementProcessor(Properties config) {
    configure(config);
  }

  // --------------- ConfigurableGenerator

  @Override
  public void configure(Properties config) {
    configTestTracing = Boolean.parseBoolean(config.getProperty(PROP_TEST_TRACING, "true"));
  }

  @Override
  public ReportPage[] getPages() {
    if (allPages.isEmpty()) {
      return new ReportPage[0];
    }

    ReportPage[] rp = new ReportPage[allPages.size() + 1];
    allPages.toArray(rp);
    rp[rp.length - 1] = new StaticResource(MainApp.CSS_SRC, MainApp.CSS);
    return rp;
  }

  // --------------- RequirementListener

  @Override
  public void attachedToAdapter(String aspect, AspectAdapter adapter) {
    this.adapter = adapter;
  }

  @Override
  public void processRequirementDb(String dbId, String dbName, String location) {
    curPage = new DBReportPage(adapter.getCurrentChannel(), dbId, dbName);
    allPages.add(curPage);
    adapter.addProcessorTag("requirement." + reqNum, null); // Hope there is only one requirement
    reqNum++;
  }

  // TODO for methods from interface
  @Override
  public void processRequirementPassed(String id, String db) {
    // TODO Auto-generated method stub
    // System.out.println("in processRequirementPassed");
  }

  @Override
  public void processRequirementFailed(String id, String db, String message) {
    // TODO Auto-generated method stub
    // curPage = new RequirementPage(adapter.getCurrentChannel(), dbName, location, message,
    // reqNum);
    // System.out.println("RequirementProcessor: message = " + message);
    // curPage.setMessage(message);
  }

  @Override
  public void processTagChange(TagMessage msg) {}

  // --------------- HtmlReportPage

  protected DBReportPage curPage;
  protected Set<DBReportPage> allPages = new LinkedHashSet<DBReportPage>();

  protected static class RequirementPage extends HtmlReportPage {

    String dbName;
    String location;
    String id;
    String message;
    String tag;

    public String getDbName() {
      return dbName;
    }

    public void setDbName(String dbName) {
      this.dbName = dbName;
    }

    public String getLocation() {
      return location;
    }

    public void setLocation(String location) {
      this.location = location;
    }

    public String getId() {
      return id;
    }

    public void setId(String id) {
      this.id = id;
    }

    public String getMessage() {
      return message;
    }

    public void setMessage(String message) {
      this.message = message;
    }

    public String getTag() {
      return tag;
    }

    public void setTag(String tag) {
      this.tag = tag;
    }

    RequirementPage(String channel, int reqNum) {
      tag = "requirement." + reqNum;
      setAssociatedChannel(channel);
      setPath("requirements/" + "req" + reqNum + ".html");
      setTitle("requirement " + reqNum);
      setStyle(MainApp.CSS);
    }

    @Override
    public void generatePageBody(PrintWriter pw) {
      pw.println("<h1>Requirement " + dbName + " from " + location + "</h1>");
      pw.println("Message: <b>" + message + "</b>");
    }

    @Override
    public String generateRef(HtmlReportPage from, String mark) {
      if (tag.equals(mark)) {
        mark = null;
      }
      return super.generateRef(from, mark);
    }
  }

  // --------------- ContextProvider

  @Override
  public HtmlReportPage[] getRelated(String channel, final HtmlReportPage target) {
    if (curPage == null) {
      return null;
    }
    final DBReportPage reqPage = curPage;
    return new HtmlReportPage[] {
      new HtmlReportPage() {
        @Override
        public void generatePageBody(PrintWriter pw) {
          pw.println("<table class=\"section\" width=\"100%\">");
          pw.println("<tr><th colspan=\"2\">Test suite");
          pw.println("<tr><td>Requirement<td>" + reqPage.generateRef(target, null));
          pw.println("</table>");
        }
      }
    };
  }

  // --------------- ReferenceResolver

  // Resolve something to this
  @Override
  public HtmlReportPage resolve(String channel, String ref) {
    return null;
  }

  @Override
  public void processRequirementStart(RequirementStartMessage msg) {
    Map<String, Set<String>> context = configTestTracing ? msg.getAllTags() : null;
    System.out.println(
        "Processing requirement start message, msg.getAllTags = " + msg.getAllTags());
    // TODO Auto-generated method stub

  }
}
