/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.tracer;

import com.unitesk.aspectrace.DebugMessage;
import com.unitesk.aspectrace.TagMessage;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;
import com.unitesk.aspectrace.Tracer;

import java.util.Map;
import java.util.Properties;

/** Formats trace into external XML-representation. */
public class XmlFormatter implements Tracer.Formatter, XmlConstants {
  private static final String nl = System.getProperty("line.separator");

  private String encoding;
  private String defChannel;

  private StringBuilder nodeSb = new StringBuilder(512);

  @Override
  public String formatTraceStart(Properties props) {
    encoding = props.getProperty(Tracer.PROP_FILE_ENCODING, Tracer.DEFAULT_ENCODING);
    defChannel = props.getProperty(Tracer.PROP_DEF_CHANNEL, Tracer.DEFAULT_CHANNEL);
    StringBuilder sb = new StringBuilder(256);

    sb.append("<?xml");
    appendAttr(sb, "version", "1.0");
    appendAttr(sb, "encoding", encoding);
    sb.append("?>").append(nl);

    sb.append('<').append(TAGNAME_TRACE);
    appendAttr(sb, "format", "mixed");
    appendAttr(sb, "version", "3.0");
    if (defChannel != null) {
      appendAttr(sb, ATTR_TRACE_DEFCHANNEL, defChannel);
    }
    sb.append('>').append(nl);
    return sb.toString();
  }

  @Override
  public synchronized String formatMessage(TraceMessage msg) {
    this.nodeNs = msg.getAspect();
    this.nodeChannel = msg.getChannel();
    appendNode(msg.getNode());
    String str = nodeSb.toString();
    nodeSb.setLength(0);
    return str;
  }

  private StringBuilder formatSpaces = new StringBuilder();
  private String nodeNs;
  private String nodeChannel;

  private void appendNode(TraceNode node) {
    if (node.isText()) {
      nodeSb.append(encodeText(node.getText()));
      return;
    }

    String fullName =
        (nodeNs == null || nodeNs.length() == 0) ? node.getName() : nodeNs + ':' + node.getName();
    nodeNs = null;

    nodeSb.append(formatSpaces);
    nodeSb.append('<').append(fullName);
    if (nodeChannel != null && !nodeChannel.equals(defChannel)) {
      appendAttr(nodeSb, ATTR_CHANNEL, nodeChannel);
    }
    nodeChannel = null;

    for (Map.Entry<String, String> me : node.getAttrs().entrySet()) {
      appendAttr(nodeSb, me.getKey(), me.getValue());
    }

    if (node.hasChildren()) {
      nodeSb.append('>');
      if (node.getChildren().size() == 1 && node.getChild(0).isText()) {
        // Special case: no linefeeds
        nodeSb.append(encodeText(node.getChild(0).getText()));
      } else {
        nodeSb.append(nl);
        for (TraceNode child : node.getChildren()) {
          formatSpaces.append("  ");
          appendNode(child);
          formatSpaces.setLength(formatSpaces.length() - 2);
        }
        nodeSb.append(formatSpaces);
      }
      nodeSb.append("</").append(fullName).append('>');
    } else {
      nodeSb.append("/>");
    }
    nodeSb.append(nl);
  }

  @Override
  public String formatTagMessage(TagMessage msg) {
    if (msg.isUser()) {
      throw new IllegalArgumentException("Tracing user tag");
    }

    StringBuilder sb = new StringBuilder(128);
    String channel = msg.getChannel();
    String aspect = msg.getAspect();
    String tag = msg.getTag();
    sb.append('<');
    if (msg.isSync()) {
      sb.append(TAGNAME_SYNC);
      appendAttr(sb, ATTR_SYNC_CHANNEL, channel);
      appendAttr(sb, ATTR_SYNC_ASPECT, aspect);
      appendAttr(sb, ATTR_SYNC_NAME, tag);
    } else {
      sb.append(msg.isStart() ? TAGNAME_TAG_START : TAGNAME_TAG_END);
      if (channel != null && !channel.equals(defChannel)) {
        appendAttr(sb, ATTR_CHANNEL, channel);
      }
      appendAttr(sb, ATTR_TAG_ASPECT, aspect);
      appendAttr(sb, ATTR_TAG_NAME, tag);
      if (msg.getParent() != null) {
        appendAttr(sb, ATTR_TAG_PARENT, msg.getParent());
      }
    }
    sb.append("/>").append(nl);
    return sb.toString();
  }

  @Override
  public String formatDebugMessage(DebugMessage msg) {
    return "<!-- " + msg.getText() + " -->" + nl;
  }

  @Override
  public String formatTraceEnd() {
    return "</" + TAGNAME_TRACE + '>' + nl;
  }

  private static void appendAttr(StringBuilder sb, String attrName, String attrValue) {
    if (attrName == null || attrValue == null) {
      return;
    }
    sb.append(' ');
    sb.append(attrName);
    sb.append("=\"");
    sb.append(encodeAttr(attrValue));
    sb.append('\"');
  }

  /**
   * Encodes an attribute value. Returns CharSequence so that the calling method would call method
   * StringBuilder.append(CharSequence) afterwards instead of StringBuilder.toString() which doubles the data.
   */
  private static CharSequence encodeAttr(String str) {
    int length = str.length();
    StringBuilder sb = new StringBuilder(length * 2 + 16);
    for (int i = 0; i < length; i++) {
      char ch = str.charAt(i);
      // TODO if (ch < ' ' && ch != '\r' && ch != '\n')...
      switch (ch) {
        case '\"':
          sb.append("&quot;");
          break;
        case '&':
          sb.append("&amp;");
          break;
        case '<':
          sb.append("&lt;");
          break;
        case '>':
          sb.append("&gt;");
          break;
        case '\r':
          if (i + 1 < str.length() && str.charAt(i + 1) == '\n') {
            break; // skip CR just before LF
          }
          sb.append("&#x0D;");
          break;
        case '\n':
          sb.append("&#x0A;");
          break;
        default:
          sb.append(ch);
      }
    }

    if (sb.length() == length && str.indexOf('\r') < 0) {
      return str;
    } else {
      return sb;
    }
  }

  /**
   * Encodes text node. Wrap in CDATA if necessary.
   *
   * @see XmlFormatter#encodeAttr(String)
   */
  private CharSequence encodeText(String str) {
    int length = str.length();
    StringBuilder sb = new StringBuilder(length * 2 + 16);
    sb.append("<![CDATA["); // length == 9
    for (int i = 0; i < length; i++) {
      char ch = str.charAt(i);
      // TODO if (ch < ' ' && ch != '\r' && ch != '\n')...
      switch (ch) {
        case '&':
          sb.append("&amp;");
          break;
        case '>':
          sb.append("&gt;");
          break;
        case '\r':
          if (i + 1 < str.length() && str.charAt(i + 1) == '\n') {
            break; // skip CR just before LF
          }
          // else fall through
        default:
          sb.append(ch);
      }
    }

    if (sb.length() == (length + 9) && str.indexOf('\r') < 0) {
      return str;
    } else {
      return sb.append("]]>");
    }
  }

  @Override
  public boolean equals(Object obj) {
    return obj != null && obj.getClass() == XmlFormatter.class;
  }

  @Override
  public int hashCode() {
    return XmlFormatter.class.hashCode();
  }
}
