/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import com.unitesk.aspectrace.coverage.CoverageTracer;
import com.unitesk.aspectrace.error.ErrorTracer;
import com.unitesk.aspectrace.html.ReportGenerator;
import com.unitesk.aspectrace.sampleApp.CoverageCollector;
import com.unitesk.aspectrace.sampleApp.TestProcessor;
import com.unitesk.aspectrace.test.TestTracer;
import com.unitesk.coverage.Coverage;
import com.unitesk.coverage.CoverageBase;
import com.unitesk.coverage.CoverageCombinator;
import com.unitesk.coverage.CoverageElement;
import com.unitesk.coverage.CoverageProcessor;
import org.junit.Test;

import java.io.File;
import java.util.HashSet;

public class MicroTeskTestCase {

  private static final String TRACE_NAME = "MicroTESK.atrace";
  private static final String TRACE_PATH = "./build/test-outputs/MicroTeskTestCase/";

  @Test
  public void test() {
    main(new String[] {});
  }

  public static void main(String[] args) {
    generateTrace();
    generateReport();
  }

  // ------------------------------------------------------

  public static interface TestCaseGenerator {
    String generateTest();
  }

  private static int testNum = 0;

  public static void doTest(TestCaseGenerator tcg) {
    testNum++;
    System.err.println("Test #" + testNum + ": " + tcg.generateTest());
  }

  @Coverage(name = "Instruction coverage: ADD")
  public enum AddSits implements TestCaseGenerator {
    Normal,
    IntegerOverflow;

    @Override
    public String generateTest() {
      return "ADD: " + name();
    }
  }

  @Coverage(name = "Instruction coverage: DIV")
  public static enum DivSits implements TestCaseGenerator {
    Normal,
    DivisionByZero;

    @Override
    public String generateTest() {
      return "DIV: " + name();
    }
  }

  @Coverage(
      aspect = "MicroTESK",
      name = "Instructions grade equivalency coverage ADD/DIV - case 1")
  public static enum AddDivSits implements TestCaseGenerator {
    Normal,
    IntegerOverflow,
    DivisionByZero;

    public static AddDivSits get(Enum<?> en)
        throws IllegalArgumentException { // Understands constants from AddSits and DivSits
      return Enum.valueOf(AddDivSits.class, en.name());
    }

    @Override
    public String generateTest() {
      switch (this) {
        case Normal:
          return "ADD: Normal";
        case IntegerOverflow:
          return "ADD: IntegerOverflow";
        case DivisionByZero:
          return "DIV: DivisionByZero";
        default:
          return null;
      }
    }
  }

  @Coverage(
      aspect = "MicroTESK",
      name = "Instructions grade equivalency coverage ADD/DIV - case 2",
      comb = CoverageCombinator.UNION)
  public static class AddDivSits2 extends CoverageBase implements TestCaseGenerator {
    public AddSits addSits;
    public DivSits divSits;

    @Override
    public String generateTest() {
      if (addSits != null) {
        return addSits.generateTest();
      } else if (divSits != null) {
        return divSits.generateTest();
      } else {
        return null;
      }
    }
  }

  @Coverage(aspect = "MicroTESK", name = "Instructions grade equivalency coverage LOAD/STORE")
  public static enum LoadStoreSits implements TestCaseGenerator {
    Normal,
    BadAddress,
    NoReadAccess,
    NoWriteAccess;

    @Override
    public String generateTest() {
      switch (this) {
        case Normal:
          return "LOAD: Normal";
        case BadAddress:
          return "STORE: BadAddress";
        case NoReadAccess:
          return "LOAD: NoReadAccess";
        case NoWriteAccess:
          return "STORE: NoWriteAccess";
        default:
          return null;
      }
    }
  }

  @Coverage
  public static enum RegDependency {
    REG_EQ,
    REG_NE;
  }

  @Coverage
  public static enum AddrDependency {
    ADDR_EQ,
    CACHE_ROW_EQ,
    CACHE_ROW_NE;
  }

  @Coverage(comb = CoverageCombinator.PRODUCT)
  public static class RegDepVector extends CoverageBase {
    public RegDependency r1;
    public RegDependency r2;
    public RegDependency r3;
    public RegDependency r4;

    @Override
    public String toString() {
      return r1.name() + "," + r2 + "," + r3 + "," + r4;
    }
  }

  @Coverage(
      aspect = "MicroTESK",
      name = "Dependency coverage for two instructions",
      comb = CoverageCombinator.PRODUCT)
  public static class TwoInstrDep extends CoverageBase implements TestCaseGenerator {
    @Coverage(prefix = "RegDeps: ", flatten = true)
    public RegDepVector r1234;

    @Coverage(prefix = "AddrDep: ")
    public AddrDependency r5;

    @Override
    public String toString() {
      return generateTest();
    }

    @Override
    public String generateTest() {
      return "TwoInstrDep: [[" + r1234.toString() + "], " + r5 + "]";
    }
  }

  @Coverage(
      aspect = "MicroTESK",
      name = "Length 3 ADD/SUB sequence instruction coverage",
      comb = CoverageCombinator.PRODUCT)
  public static class InstrSeq3 extends CoverageBase implements TestCaseGenerator {
    @Coverage(flatten = true)
    public AddDivSits sit1;

    @Coverage(flatten = true)
    public TwoInstrDep d12; // +

    @Coverage(flatten = true)
    public AddDivSits sit2;

    @Coverage(flatten = true)
    public TwoInstrDep d13; // +

    @Coverage(flatten = true)
    public AddDivSits sit3;

    @Override
    public String generateTest() {
      return "InstrSeq3 [sit1="
          + sit1
          + ", sit2="
          + sit2
          + ", sit3="
          + sit3
          + ", d12="
          + d12
          + ", d13="
          + d13
          + "]";
    }
  }

  // Random filtration/aggregation of another coverage structure for report.
  @Coverage(
      aspect = "MicroTESK",
      name = "Length 3 ADD/SUB sequence instruction coverage (filtered)",
      comb = CoverageCombinator.PRODUCT)
  public static class InstrSeq3Filtered extends CoverageBase {
    @Coverage(flatten = true)
    public TwoInstrDep d12;

    @Coverage(flatten = true)
    public TwoInstrDep d13;

    public InstrSeq3Filtered(InstrSeq3 src) {
      d12 = (TwoInstrDep) src.d12; // .clone();
      d13 = (TwoInstrDep) src.d13; // .clone();
    }
  }

  // ------------------------------------------------------

  public static void generateTrace() {
    boolean verdict = true;
    try {
      System.err.println(">>> Generating trace");

      Tracer.getInstance().addXmlTrace(new File(TRACE_PATH + TRACE_NAME));
      TestTracer.traceTestStart("MicroTESK");

      CoverageTracer.traceCoverageStructure(AddSits.class);
      CoverageTracer.traceCoverageStructure(DivSits.class);
      CoverageTracer.traceCoverageStructure(AddDivSits.class);
      CoverageTracer.traceCoverageStructure(AddDivSits2.class);
      CoverageTracer.traceCoverageStructure(TwoInstrDep.class);
      CoverageTracer.traceCoverageStructure(InstrSeq3Filtered.class);

      /* Test situation enumeration. */

      // Reports about several coverages at once.
      for (AddSits as : CoverageProcessor.enumerator(AddSits.class)) {
        doTest(as);
        CoverageTracer.traceCoverageElement(as);
        CoverageTracer.traceCoverageElement(AddDivSits.get(as));
      }
      for (DivSits ds : CoverageProcessor.enumerator(DivSits.class)) {
        doTest(ds);
        CoverageTracer.traceCoverageElement(ds);
        CoverageTracer.traceCoverageElement(AddDivSits.get(ds));
      }
      // Report what we are looking through
      for (AddDivSits2 ads : CoverageProcessor.enumerator(AddDivSits2.class)) {
        doTest(ads);
        CoverageTracer.traceCoverageElement(ads);
      }
      for (TwoInstrDep tid : CoverageProcessor.enumerator(TwoInstrDep.class)) {
        doTest(tid);
        CoverageTracer.traceCoverageElement(tid);
      }

      // Filtered coverage report. TOTAL CHECK of all test cases.
      // However, not all information is in report.
      long count = CoverageProcessor.countValues(InstrSeq3.class);
      {
        HashSet<CoverageElement> covered =
            new HashSet<CoverageElement>(
                (int) CoverageProcessor.countValues(InstrSeq3Filtered.class));
        for (InstrSeq3 is3 : CoverageProcessor.enumerator(InstrSeq3.class)) {
          System.err.print(count + "// ");
          doTest(is3);

          // Reduces volume of test case information and number of TC.
          InstrSeq3Filtered filter = new InstrSeq3Filtered(is3);
          CoverageElement elem = filter.getElement();
          // Reduces trace volume: reports each TC only once.
          if (covered.add(elem)) {
            CoverageTracer.traceCoverageElement(elem);
            System.err.println("-> traced");
          } else {
            System.err.println("-> already exists");
          }
        }
      }

    } catch (Throwable th) {
      th.printStackTrace();
      verdict = false;
      ErrorTracer.traceFailure("test", "Exception", null, th);
    } finally {
      TestTracer.traceTestEnd("MicroTESK", verdict);
      // Ends a trace.
      Tracer.getInstance().endTrace();
    }
  }

  public static void generateReport() {
    try {
      System.err.println(">>> Generating report");
      TraceProcessorManager tpm = new TraceProcessorManager();
      ReportGenerator rg = new ReportGenerator();
      rg.setDefaultCharset("windows-1251");
      // Adds input trace.
      tpm.addTrace(new File(TRACE_PATH + TRACE_NAME));
      // Adds component which handles coverage structures.
      CoverageCollector cc = new CoverageCollector();
      tpm.addListener("coverage", cc);
      rg.addRefResolver("coverage", cc);
      // Adds component which handles tests.
      TestProcessor tp = new TestProcessor();
      tpm.addListener("test", tp);
      rg.addRefResolver("test", tp);
      rg.addAutoRef("coverage", "test");
      // Starts processing.
      try {
        tpm.process();
      } catch (Exception ex) {
        ex.printStackTrace();
        System.err.println();
        System.err.println("Trying to generate report after failure...");
      }
      ////////// Makes an HTML-report. //////////
      rg.setGenerationDir(new File("./build/test-outputs/MicroTeskTestCase/"));
      rg.addResource("/com/unitesk/aspectrace/sampleApp/report.css", "report.css");
      rg.addIndexPage();
      rg.addPages(cc.getPages());
      rg.addPages(tp.getPages());
      rg.generateReport();
    } catch (Throwable th) {
      th.printStackTrace();
    }
  }
}
